@@ -1,5 +1,5 @@
 /*
- * $PostgreSQL: pgsql/contrib/pgbench/pgbench.c,v 1.45.2.5 2007/01/10 01:44:30 ishii Exp $
+ * $PostgreSQL: pgsql/contrib/pgbench/pgbench.c,v 1.58.2.2 2007/08/22 23:03:33 tgl Exp $
  *
  * pgbench: a simple benchmark program for PostgreSQL
  * written by Tatsuo Ishii
@@ -37,8 +37,9 @@
 #include <sys/select.h>
 #endif
 
-/* for getrlimit */
-#include <sys/resource.h>
+#ifdef HAVE_SYS_RESOURCE_H
+#include <sys/resource.h>		/* for getrlimit */
+#endif
 #endif   /* ! WIN32 */
 
 extern char *optarg;
@@ -58,10 +59,10 @@ int			nclients = 1;		/* default number of simulated clients */
 int			nxacts = 10;		/* default number of transactions per clients */
 
 /*
- * scaling factor. for example, tps = 10 will make 1000000 tuples of
+ * scaling factor. for example, scale = 10 will make 1000000 tuples of
  * accounts table.
  */
-int			tps = 1;
+int			scale = 1;
 
 /*
  * end of configurable parameters
@@ -80,7 +81,7 @@ int			remains;			/* number of remaining clients */
 int			is_connect;			/* establish connection  for each transaction */
 
 char	   *pghost = "";
-char	   *pgport = NULL;
+char	   *pgport = "";
 char	   *pgoptions = NULL;
 char	   *pgtty = NULL;
 char	   *login = NULL;
@@ -134,10 +135,13 @@ int			num_files;			/* its number */
 
 /* default scenario */
 static char *tpc_b = {
-	"\\setrandom aid 1 100000\n"
-	"\\setrandom bid 1 1\n"
-	"\\setrandom tid 1 10\n"
-	"\\setrandom delta 1 10000\n"
+	"\\set nbranches :scale\n"
+	"\\set ntellers 10 * :scale\n"
+	"\\set naccounts 100000 * :scale\n"
+	"\\setrandom aid 1 :naccounts\n"
+	"\\setrandom bid 1 :nbranches\n"
+	"\\setrandom tid 1 :ntellers\n"
+	"\\setrandom delta -5000 5000\n"
 	"BEGIN;\n"
 	"UPDATE accounts SET abalance = abalance + :delta WHERE aid = :aid;\n"
 	"SELECT abalance FROM accounts WHERE aid = :aid;\n"
@@ -149,10 +153,13 @@ static char *tpc_b = {
 
 /* -N case */
 static char *simple_update = {
-	"\\setrandom aid 1 100000\n"
-	"\\setrandom bid 1 1\n"
-	"\\setrandom tid 1 10\n"
-	"\\setrandom delta 1 10000\n"
+	"\\set nbranches :scale\n"
+	"\\set ntellers 10 * :scale\n"
+	"\\set naccounts 100000 * :scale\n"
+	"\\setrandom aid 1 :naccounts\n"
+	"\\setrandom bid 1 :nbranches\n"
+	"\\setrandom tid 1 :ntellers\n"
+	"\\setrandom delta -5000 5000\n"
 	"BEGIN;\n"
 	"UPDATE accounts SET abalance = abalance + :delta WHERE aid = :aid;\n"
 	"SELECT abalance FROM accounts WHERE aid = :aid;\n"
@@ -162,14 +169,15 @@ static char *simple_update = {
 
 /* -S case */
 static char *select_only = {
-	"\\setrandom aid 1 100000\n"
+	"\\set naccounts 100000 * :scale\n"
+	"\\setrandom aid 1 :naccounts\n"
 	"SELECT abalance FROM accounts WHERE aid = :aid;\n"
 };
 
 static void
 usage(void)
 {
-	fprintf(stderr, "usage: pgbench [-h hostname][-p port][-c nclients][-t ntransactions][-s scaling_factor][-n][-C][-v][-S][-N][-f filename][-l][-U login][-P password][-d][dbname]\n");
+	fprintf(stderr, "usage: pgbench [-h hostname][-p port][-c nclients][-t ntransactions][-s scaling_factor][-D varname=value][-n][-C][-v][-S][-N][-f filename][-l][-U login][-P password][-d][dbname]\n");
 	fprintf(stderr, "(initialize mode): pgbench -i [-h hostname][-p port][-s scaling_factor][-U login][-P password][-d][dbname]\n");
 }
 
@@ -331,10 +339,13 @@ putVariable(CState * st, char *name, char *value)
 	}
 	else
 	{
-		if ((value = strdup(value)) == NULL)
+		char	   *val;
+
+		if ((val = strdup(value)) == NULL)
 			return false;
+
 		free(var->value);
-		var->value = value;
+		var->value = val;
 	}
 
 	return true;
@@ -450,7 +461,7 @@ top:
 		if (commands[st->state]->type == SQL_COMMAND)
 		{
 			res = PQgetResult(st->con);
-			if (strncasecmp(commands[st->state]->argv[0], "select", 6) != 0)
+			if (pg_strncasecmp(commands[st->state]->argv[0], "select", 6) != 0)
 			{
 				if (check(state, res, n, PGRES_COMMAND_OK))
 					return;
@@ -550,28 +561,137 @@ top:
 			fprintf(stderr, "\n");
 		}
 
-		if (strcasecmp(argv[0], "setrandom") == 0)
+		if (pg_strcasecmp(argv[0], "setrandom") == 0)
 		{
-			char	   *val;
+			char	   *var;
+			int			min,
+						max;
+			char		res[64];
 
-			if ((val = malloc(strlen(argv[3]) + 1)) == NULL)
+			if (*argv[2] == ':')
+			{
+				if ((var = getVariable(st, argv[2] + 1)) == NULL)
+				{
+					fprintf(stderr, "%s: undefined variable %s\n", argv[0], argv[2]);
+					st->ecnt++;
+					return;
+				}
+				min = atoi(var);
+			}
+			else
+				min = atoi(argv[2]);
+
+#ifdef NOT_USED
+			if (min < 0)
+			{
+				fprintf(stderr, "%s: invalid minimum number %d\n", argv[0], min);
+				st->ecnt++;
+				return;
+			}
+#endif
+
+			if (*argv[3] == ':')
+			{
+				if ((var = getVariable(st, argv[3] + 1)) == NULL)
+				{
+					fprintf(stderr, "%s: undefined variable %s\n", argv[0], argv[3]);
+					st->ecnt++;
+					return;
+				}
+				max = atoi(var);
+			}
+			else
+				max = atoi(argv[3]);
+
+			if (max < min || max > MAX_RANDOM_VALUE)
+			{
+				fprintf(stderr, "%s: invalid maximum number %d\n", argv[0], max);
+				st->ecnt++;
+				return;
+			}
+
+#ifdef DEBUG
+			printf("min: %d max: %d random: %d\n", min, max, getrand(min, max));
+#endif
+			snprintf(res, sizeof(res), "%d", getrand(min, max));
+
+			if (putVariable(st, argv[1], res) == false)
 			{
 				fprintf(stderr, "%s: out of memory\n", argv[0]);
 				st->ecnt++;
 				return;
 			}
 
-			sprintf(val, "%d", getrand(atoi(argv[2]), atoi(argv[3])));
+			st->listen = 1;
+		}
+		else if (pg_strcasecmp(argv[0], "set") == 0)
+		{
+			char	   *var;
+			int			ope1,
+						ope2;
+			char		res[64];
 
-			if (putVariable(st, argv[1], val) == false)
+			if (*argv[2] == ':')
+			{
+				if ((var = getVariable(st, argv[2] + 1)) == NULL)
+				{
+					fprintf(stderr, "%s: undefined variable %s\n", argv[0], argv[2]);
+					st->ecnt++;
+					return;
+				}
+				ope1 = atoi(var);
+			}
+			else
+				ope1 = atoi(argv[2]);
+
+			if (argc < 5)
+				snprintf(res, sizeof(res), "%d", ope1);
+			else
+			{
+				if (*argv[4] == ':')
+				{
+					if ((var = getVariable(st, argv[4] + 1)) == NULL)
+					{
+						fprintf(stderr, "%s: undefined variable %s\n", argv[0], argv[4]);
+						st->ecnt++;
+						return;
+					}
+					ope2 = atoi(var);
+				}
+				else
+					ope2 = atoi(argv[4]);
+
+				if (strcmp(argv[3], "+") == 0)
+					snprintf(res, sizeof(res), "%d", ope1 + ope2);
+				else if (strcmp(argv[3], "-") == 0)
+					snprintf(res, sizeof(res), "%d", ope1 - ope2);
+				else if (strcmp(argv[3], "*") == 0)
+					snprintf(res, sizeof(res), "%d", ope1 * ope2);
+				else if (strcmp(argv[3], "/") == 0)
+				{
+					if (ope2 == 0)
+					{
+						fprintf(stderr, "%s: division by zero\n", argv[0]);
+						st->ecnt++;
+						return;
+					}
+					snprintf(res, sizeof(res), "%d", ope1 / ope2);
+				}
+				else
+				{
+					fprintf(stderr, "%s: unsupported operator %s\n", argv[0], argv[3]);
+					st->ecnt++;
+					return;
+				}
+			}
+
+			if (putVariable(st, argv[1], res) == false)
 			{
 				fprintf(stderr, "%s: out of memory\n", argv[0]);
-				free(val);
 				st->ecnt++;
 				return;
 			}
 
-			free(val);
 			st->listen = 1;
 		}
 
@@ -639,7 +759,7 @@ init(void)
 	}
 	PQclear(res);
 
-	for (i = 0; i < nbranches * tps; i++)
+	for (i = 0; i < nbranches * scale; i++)
 	{
 		snprintf(sql, 256, "insert into branches(bid,bbalance) values(%d,0)", i + 1);
 		res = PQexec(con, sql);
@@ -651,7 +771,7 @@ init(void)
 		PQclear(res);
 	}
 
-	for (i = 0; i < ntellers * tps; i++)
+	for (i = 0; i < ntellers * scale; i++)
 	{
 		snprintf(sql, 256, "insert into tellers(tid,bid,tbalance) values (%d,%d,0)"
 				 ,i + 1, i / ntellers + 1);
@@ -676,7 +796,7 @@ init(void)
 	 * occupy accounts table with some data
 	 */
 	fprintf(stderr, "creating tables...\n");
-	for (i = 0; i < naccounts * tps; i++)
+	for (i = 0; i < naccounts * scale; i++)
 	{
 		int			j = i + 1;
 
@@ -806,11 +926,8 @@ process_commands(char *buf)
 			tok = strtok(NULL, delim);
 		}
 
-		if (strcasecmp(my_commands->argv[0], "setrandom") == 0)
+		if (pg_strcasecmp(my_commands->argv[0], "setrandom") == 0)
 		{
-			int			min,
-						max;
-
 			if (my_commands->argc < 4)
 			{
 				fprintf(stderr, "%s: missing argument\n", my_commands->argv[0]);
@@ -820,20 +937,18 @@ process_commands(char *buf)
 			for (j = 4; j < my_commands->argc; j++)
 				fprintf(stderr, "%s: extra argument \"%s\" ignored\n",
 						my_commands->argv[0], my_commands->argv[j]);
-
-			if ((min = atoi(my_commands->argv[2])) < 0)
+		}
+		else if (pg_strcasecmp(my_commands->argv[0], "set") == 0)
+		{
+			if (my_commands->argc < 3)
 			{
-				fprintf(stderr, "%s: invalid minimum number %s\n",
-						my_commands->argv[0], my_commands->argv[2]);
+				fprintf(stderr, "%s: missing argument\n", my_commands->argv[0]);
 				return NULL;
 			}
 
-			if ((max = atoi(my_commands->argv[3])) < min || max > MAX_RANDOM_VALUE)
-			{
-				fprintf(stderr, "%s: invalid maximum number %s\n",
-						my_commands->argv[0], my_commands->argv[3]);
-				return NULL;
-			}
+			for (j = my_commands->argc < 5 ? 3 : 5; j < my_commands->argc; j++)
+				fprintf(stderr, "%s: extra argument \"%s\" ignored\n",
+						my_commands->argv[0], my_commands->argv[j]);
 		}
 		else
 		{
@@ -889,16 +1004,22 @@ process_file(char *filename)
 	while (fgets(buf, sizeof(buf), fd) != NULL)
 	{
 		Command    *commands;
+		int			i;
 
+		i = 0;
+		while (isspace((unsigned char) buf[i]))
+			i++;
 
-		if (strncmp(buf, "\n", 1) != 0) {
-			commands = process_commands(buf);
+		if (buf[i] != '\0' && strncmp(&buf[i], "--", 2) != 0)
+		{
+			commands = process_commands(&buf[i]);
 			if (commands == NULL)
 			{
 				fclose(fd);
 				return false;
 			}
-		} else
+		}
+		else
 			continue;
 
 		my_commands[lineno] = commands;
@@ -1018,7 +1139,7 @@ printResults(
 		s = "Custom query";
 
 	printf("transaction type: %s\n", s);
-	printf("scaling factor: %d\n", tps);
+	printf("scaling factor: %d\n", scale);
 	printf("number of clients: %d\n", nclients);
 	printf("number of transactions per client: %d\n", nxacts);
 	printf("number of transactions actually processed: %d/%d\n", normal_xacts, nxacts * nclients);
@@ -1052,7 +1173,7 @@ main(int argc, char **argv)
 	int			nsocks;			/* return from select(2) */
 	int			maxsock;		/* max socket number to be waited */
 
-#if !(defined(__CYGWIN__) || defined(__MINGW32__))
+#ifdef HAVE_GETRLIMIT
 	struct rlimit rlim;
 #endif
 
@@ -1060,6 +1181,8 @@ main(int argc, char **argv)
 	PGresult   *res;
 	char	   *env;
 
+	char		val[64];
+
 	if ((env = getenv("PGHOST")) != NULL && *env != '\0')
 		pghost = env;
 	if ((env = getenv("PGPORT")) != NULL && *env != '\0')
@@ -1067,7 +1190,16 @@ main(int argc, char **argv)
 	else if ((env = getenv("PGUSER")) != NULL && *env != '\0')
 		login = env;
 
-	while ((c = getopt(argc, argv, "ih:nvp:dc:t:s:U:P:CNSlf:")) != -1)
+	state = (CState *) malloc(sizeof(CState));
+	if (state == NULL)
+	{
+		fprintf(stderr, "Couldn't allocate memory for state\n");
+		exit(1);
+	}
+
+	memset(state, 0, sizeof(*state));
+
+	while ((c = getopt(argc, argv, "ih:nvp:dc:t:s:U:P:CNSlf:D:")) != -1)
 	{
 		switch (c)
 		{
@@ -1102,8 +1234,8 @@ main(int argc, char **argv)
 					fprintf(stderr, "invalid number of clients: %d\n", nclients);
 					exit(1);
 				}
-#if !(defined(__CYGWIN__) || defined(__MINGW32__))
-#ifdef RLIMIT_NOFILE			/* most platform uses RLIMIT_NOFILE */
+#ifdef HAVE_GETRLIMIT
+#ifdef RLIMIT_NOFILE			/* most platforms use RLIMIT_NOFILE */
 				if (getrlimit(RLIMIT_NOFILE, &rlim) == -1)
 #else							/* but BSD doesn't ... */
 				if (getrlimit(RLIMIT_OFILE, &rlim) == -1)
@@ -1114,20 +1246,20 @@ main(int argc, char **argv)
 				}
 				if (rlim.rlim_cur <= (nclients + 2))
 				{
-					fprintf(stderr, "You need at least %d open files resource but you are only allowed to use %ld.\n", nclients + 2, (long) rlim.rlim_cur);
-					fprintf(stderr, "Use limit/ulimt to increase the limit before using pgbench.\n");
+					fprintf(stderr, "You need at least %d open files but you are only allowed to use %ld.\n", nclients + 2, (long) rlim.rlim_cur);
+					fprintf(stderr, "Use limit/ulimit to increase the limit before using pgbench.\n");
 					exit(1);
 				}
-#endif
+#endif /* HAVE_GETRLIMIT */
 				break;
 			case 'C':
 				is_connect = 1;
 				break;
 			case 's':
-				tps = atoi(optarg);
-				if (tps <= 0)
+				scale = atoi(optarg);
+				if (scale <= 0)
 				{
-					fprintf(stderr, "invalid scaling factor: %d\n", tps);
+					fprintf(stderr, "invalid scaling factor: %d\n", scale);
 					exit(1);
 				}
 				break;
@@ -1151,9 +1283,27 @@ main(int argc, char **argv)
 			case 'f':
 				ttype = 3;
 				filename = optarg;
-				if (process_file(filename) == false)
+				if (process_file(filename) == false || *sql_files[num_files - 1] == NULL)
 					exit(1);
 				break;
+			case 'D':
+				{
+					char	   *p;
+
+					if ((p = strchr(optarg, '=')) == NULL || p == optarg || *(p + 1) == '\0')
+					{
+						fprintf(stderr, "invalid variable definition: %s\n", optarg);
+						exit(1);
+					}
+
+					*p++ = '\0';
+					if (putVariable(&state[0], optarg, p) == false)
+					{
+						fprintf(stderr, "Couldn't allocate memory for variable\n");
+						exit(1);
+					}
+				}
+				break;
 			default:
 				usage();
 				exit(1);
@@ -1181,14 +1331,49 @@ main(int argc, char **argv)
 
 	remains = nclients;
 
-	state = (CState *) malloc(sizeof(CState) * nclients);
-	if (state == NULL)
+	if (getVariable(&state[0], "scale") == NULL)
 	{
-		fprintf(stderr, "Couldn't allocate memory for state\n");
-		exit(1);
+		snprintf(val, sizeof(val), "%d", scale);
+		if (putVariable(&state[0], "scale", val) == false)
+		{
+			fprintf(stderr, "Couldn't allocate memory for variable\n");
+			exit(1);
+		}
 	}
 
-	memset(state, 0, sizeof(*state) * nclients);
+	if (nclients > 1)
+	{
+		state = (CState *) realloc(state, sizeof(CState) * nclients);
+		if (state == NULL)
+		{
+			fprintf(stderr, "Couldn't allocate memory for state\n");
+			exit(1);
+		}
+
+		memset(state + 1, 0, sizeof(*state) * (nclients - 1));
+
+		snprintf(val, sizeof(val), "%d", scale);
+
+		for (i = 1; i < nclients; i++)
+		{
+			int			j;
+
+			for (j = 0; j < state[0].nvariables; j++)
+			{
+				if (putVariable(&state[i], state[0].variables[j].name, state[0].variables[j].value) == false)
+				{
+					fprintf(stderr, "Couldn't allocate memory for variable\n");
+					exit(1);
+				}
+			}
+
+			if (putVariable(&state[i], "scale", val) == false)
+			{
+				fprintf(stderr, "Couldn't allocate memory for variable\n");
+				exit(1);
+			}
+		}
+	}
 
 	if (use_log)
 	{
@@ -1234,13 +1419,32 @@ main(int argc, char **argv)
 			fprintf(stderr, "%s", PQerrorMessage(con));
 			exit(1);
 		}
-		tps = atoi(PQgetvalue(res, 0, 0));
-		if (tps < 0)
+		scale = atoi(PQgetvalue(res, 0, 0));
+		if (scale < 0)
 		{
-			fprintf(stderr, "count(*) from branches invalid (%d)\n", tps);
+			fprintf(stderr, "count(*) from branches invalid (%d)\n", scale);
 			exit(1);
 		}
 		PQclear(res);
+
+		snprintf(val, sizeof(val), "%d", scale);
+		if (putVariable(&state[0], "scale", val) == false)
+		{
+			fprintf(stderr, "Couldn't allocate memory for variable\n");
+			exit(1);
+		}
+
+		if (nclients > 1)
+		{
+			for (i = 1; i < nclients; i++)
+			{
+				if (putVariable(&state[i], "scale", val) == false)
+				{
+					fprintf(stderr, "Couldn't allocate memory for variable\n");
+					exit(1);
+				}
+			}
+		}
 	}
 
 	if (!is_no_vacuum)
@@ -1318,38 +1522,21 @@ main(int argc, char **argv)
 	/* process bultin SQL scripts */
 	switch (ttype)
 	{
-			char		buf[128];
-
 		case 0:
 			sql_files[0] = process_builtin(tpc_b);
-			snprintf(buf, sizeof(buf), "%d", 100000 * tps);
-			sql_files[0][0]->argv[3] = strdup(buf);
-			snprintf(buf, sizeof(buf), "%d", 1 * tps);
-			sql_files[0][1]->argv[3] = strdup(buf);
-			snprintf(buf, sizeof(buf), "%d", 10 * tps);
-			sql_files[0][2]->argv[3] = strdup(buf);
-			snprintf(buf, sizeof(buf), "%d", 10000 * tps);
-			sql_files[0][3]->argv[3] = strdup(buf);
 			num_files = 1;
 			break;
+
 		case 1:
 			sql_files[0] = process_builtin(select_only);
-			snprintf(buf, sizeof(buf), "%d", 100000 * tps);
-			sql_files[0][0]->argv[3] = strdup(buf);
 			num_files = 1;
 			break;
+
 		case 2:
 			sql_files[0] = process_builtin(simple_update);
-			snprintf(buf, sizeof(buf), "%d", 100000 * tps);
-			sql_files[0][0]->argv[3] = strdup(buf);
-			snprintf(buf, sizeof(buf), "%d", 1 * tps);
-			sql_files[0][1]->argv[3] = strdup(buf);
-			snprintf(buf, sizeof(buf), "%d", 10 * tps);
-			sql_files[0][2]->argv[3] = strdup(buf);
-			snprintf(buf, sizeof(buf), "%d", 10000 * tps);
-			sql_files[0][3]->argv[3] = strdup(buf);
 			num_files = 1;
 			break;
+
 		default:
 			break;
 	}
@@ -1357,8 +1544,19 @@ main(int argc, char **argv)
 	/* send start up queries in async manner */
 	for (i = 0; i < nclients; i++)
 	{
+		Command   **commands = sql_files[state[i].use_file];
+		int			prev_ecnt = state[i].ecnt;
+
 		state[i].use_file = getrand(0, num_files - 1);
 		doCustom(state, i, debug);
+
+		if (state[i].ecnt > prev_ecnt && commands[state[i].state]->type == META_COMMAND)
+		{
+			fprintf(stderr, "Client %d aborted in state %d. Execution meta-command failed.\n", i, state[i].state);
+			remains--;			/* I've aborted */
+			PQfinish(state[i].con);
+			state[i].con = NULL;
+		}
 	}
 
 	for (;;)
@@ -1424,12 +1622,21 @@ main(int argc, char **argv)
 		for (i = 0; i < nclients; i++)
 		{
 			Command   **commands = sql_files[state[i].use_file];
+			int			prev_ecnt = state[i].ecnt;
 
 			if (state[i].con && (FD_ISSET(PQsocket(state[i].con), &input_mask)
 						  || commands[state[i].state]->type == META_COMMAND))
 			{
 				doCustom(state, i, debug);
 			}
+
+			if (state[i].ecnt > prev_ecnt && commands[state[i].state]->type == META_COMMAND)
+			{
+				fprintf(stderr, "Client %d aborted in state %d. Execution meta-command failed.\n", i, state[i].state);
+				remains--;		/* I've aborted */
+				PQfinish(state[i].con);
+				state[i].con = NULL;
+			}
 		}
 	}
 }
