@@ -20,6 +20,8 @@
 #include "miscadmin.h"
 #include "utils/memutils.h"
 
+const XLogRecPtr XLogRecPtrForTemp = {1, 1};
+
 /* Working state for gistbuild and its callback */
 typedef struct
 {
@@ -128,7 +130,7 @@ gistbuild(PG_FUNCTION_ARGS)
 		PageSetTLI(page, ThisTimeLineID);
 	}
 	else
-		PageSetLSN(page, GetXLogRecPtrForTemp());
+		PageSetLSN(page, XLogRecPtrForTemp);
 
 	UnlockReleaseBuffer(buffer);
 
@@ -419,7 +421,7 @@ gistplacetopage(GISTInsertState *state, GISTSTATE *giststate)
 		{
 			for (ptr = dist; ptr; ptr = ptr->next)
 			{
-				PageSetLSN(ptr->page, GetXLogRecPtrForTemp());
+				PageSetLSN(ptr->page, XLogRecPtrForTemp);
 			}
 		}
 
@@ -487,7 +489,7 @@ gistplacetopage(GISTInsertState *state, GISTSTATE *giststate)
 			PageSetTLI(state->stack->page, ThisTimeLineID);
 		}
 		else
-			PageSetLSN(state->stack->page, GetXLogRecPtrForTemp());
+			PageSetLSN(state->stack->page, XLogRecPtrForTemp);
 
 		if (state->stack->blkno == GIST_ROOT_BLKNO)
 			state->needInsertComplete = false;
@@ -674,12 +676,9 @@ gistFindPath(Relation r, BlockNumber child)
 
 		if (GistPageIsLeaf(page))
 		{
-			/*
-			 * Because we scan the index top-down, all the rest of the pages
-			 * in the queue must be leaf pages as well.
-			 */
+			/* we can safety go away, follows only leaf pages */
 			UnlockReleaseBuffer(buffer);
-			break;
+			return NULL;
 		}
 
 		top->lsn = PageGetLSN(page);
@@ -687,25 +686,14 @@ gistFindPath(Relation r, BlockNumber child)
 		if (top->parent && XLByteLT(top->parent->lsn, GistPageGetOpaque(page)->nsn) &&
 			GistPageGetOpaque(page)->rightlink != InvalidBlockNumber /* sanity check */ )
 		{
-			/*
-			 * Page was split while we looked elsewhere. We didn't see the
-			 * downlink to the right page when we scanned the parent, so
-			 * add it to the queue now.
-			 *
-			 * Put the right page ahead of the queue, so that we visit it
-			 * next. That's important, because if this is the lowest internal
-			 * level, just above leaves, we might already have queued up some
-			 * leaf pages, and we assume that there can't be any non-leaf
-			 * pages behind leaf pages.
-			 */
+			/* page splited while we thinking of... */
 			ptr = (GISTInsertStack *) palloc0(sizeof(GISTInsertStack));
 			ptr->blkno = GistPageGetOpaque(page)->rightlink;
 			ptr->childoffnum = InvalidOffsetNumber;
-			ptr->parent = top->parent;
-			ptr->next = top->next;
-			top->next = ptr;
-			if (tail == top)
-				tail = ptr;
+			ptr->parent = top;
+			ptr->next = NULL;
+			tail->next = ptr;
+			tail = ptr;
 		}
 
 		maxoff = PageGetMaxOffsetNumber(page);
@@ -763,9 +751,7 @@ gistFindPath(Relation r, BlockNumber child)
 		top = top->next;
 	}
 
-	elog(ERROR, "failed to re-find parent of a page in index \"%s\", block %u",
-		 RelationGetRelationName(r), child);
-	return NULL; /* keep compiler quiet */
+	return NULL;
 }
 
 
@@ -837,6 +823,7 @@ gistFindCorrectParent(Relation r, GISTInsertStack *child)
 
 		/* ok, find new path */
 		ptr = parent = gistFindPath(r, child->blkno);
+		Assert(ptr != NULL);
 
 		/* read all buffers as expected by caller */
 		/* note we don't lock them or gistcheckpage them here! */
@@ -1038,7 +1025,7 @@ gistnewroot(Relation r, Buffer buffer, IndexTuple *itup, int len, ItemPointer ke
 		PageSetTLI(page, ThisTimeLineID);
 	}
 	else
-		PageSetLSN(page, GetXLogRecPtrForTemp());
+		PageSetLSN(page, XLogRecPtrForTemp);
 
 	END_CRIT_SECTION();
 }
