@@ -6,19 +6,18 @@
  * This gives R-tree behavior, with Guttman's poly-time split algorithm.
  *
  *
- * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2006, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
- *	$PostgreSQL: pgsql/src/backend/access/gist/gistproc.c,v 1.3.2.2 2009/04/07 17:46:37 teodor Exp $
+ *	$PostgreSQL: pgsql/src/backend/access/gist/gistproc.c,v 1.9.2.3 2009/09/18 14:03:39 teodor Exp $
  *
  *-------------------------------------------------------------------------
  */
 #include "postgres.h"
 
 #include "access/gist.h"
-#include "access/itup.h"
-#include "access/rtree.h"
+#include "access/skey.h"
 #include "utils/geo_decls.h"
 
 
@@ -33,6 +32,47 @@ static bool rtree_internal_consistent(BOX *key, BOX *query,
  * Box ops
  **************************************************/
 
+static Datum
+rt_box_union(PG_FUNCTION_ARGS)
+{
+	BOX		   *a = PG_GETARG_BOX_P(0);
+	BOX		   *b = PG_GETARG_BOX_P(1);
+	BOX		   *n;
+
+	n = (BOX *) palloc(sizeof(BOX));
+
+	n->high.x = Max(a->high.x, b->high.x);
+	n->high.y = Max(a->high.y, b->high.y);
+	n->low.x = Min(a->low.x, b->low.x);
+	n->low.y = Min(a->low.y, b->low.y);
+
+	PG_RETURN_BOX_P(n);
+}
+
+static Datum
+rt_box_inter(PG_FUNCTION_ARGS)
+{
+	BOX		   *a = PG_GETARG_BOX_P(0);
+	BOX		   *b = PG_GETARG_BOX_P(1);
+	BOX		   *n;
+
+	n = (BOX *) palloc(sizeof(BOX));
+
+	n->high.x = Min(a->high.x, b->high.x);
+	n->high.y = Min(a->high.y, b->high.y);
+	n->low.x = Max(a->low.x, b->low.x);
+	n->low.y = Max(a->low.y, b->low.y);
+
+	if (n->high.x < n->low.x || n->high.y < n->low.y)
+	{
+		pfree(n);
+		/* Indicate "no intersection" by returning NULL pointer */
+		n = NULL;
+	}
+
+	PG_RETURN_BOX_P(n);
+}
+
 /*
  * The GiST Consistent method for boxes
  *
@@ -64,6 +104,18 @@ gist_box_consistent(PG_FUNCTION_ARGS)
 												 strategy));
 }
 
+static void
+adjustBox(BOX *b, BOX *addon)
+{
+	if (b->high.x < addon->high.x)
+		b->high.x = addon->high.x;
+	if (b->low.x > addon->low.x)
+		b->low.x = addon->low.x;
+	if (b->high.y < addon->high.y)
+		b->high.y = addon->high.y;
+	if (b->low.y > addon->low.y)
+		b->low.y = addon->low.y;
+}
 
 /*
  * The GiST Union method for boxes
@@ -88,14 +140,7 @@ gist_box_union(PG_FUNCTION_ARGS)
 	for (i = 1; i < numranges; i++)
 	{
 		cur = DatumGetBoxP(entryvec->vector[i].key);
-		if (pageunion->high.x < cur->high.x)
-			pageunion->high.x = cur->high.x;
-		if (pageunion->low.x > cur->low.x)
-			pageunion->low.x = cur->low.x;
-		if (pageunion->high.y < cur->high.y)
-			pageunion->high.y = cur->high.y;
-		if (pageunion->low.y > cur->low.y)
-			pageunion->low.y = cur->low.y;
+		adjustBox(pageunion, cur);
 	}
 	*sizep = sizeof(BOX);
 
@@ -143,16 +188,83 @@ gist_box_penalty(PG_FUNCTION_ARGS)
 }
 
 static void
-adjustBox(BOX *b, BOX *addon)
+chooseLR(GIST_SPLITVEC *v,
+		 OffsetNumber *list1, int nlist1, BOX *union1,
+		 OffsetNumber *list2, int nlist2, BOX *union2)
 {
-    if (b->high.x < addon->high.x)
-		b->high.x = addon->high.x;
-	if (b->low.x > addon->low.x)
-		b->low.x = addon->low.x;
-	if (b->high.y < addon->high.y)
-		b->high.y = addon->high.y;
-	if (b->low.y > addon->low.y)
-		b->low.y = addon->low.y;
+	bool		firstToLeft = true;
+
+	if (v->spl_ldatum_exists || v->spl_rdatum_exists)
+	{
+		if (v->spl_ldatum_exists && v->spl_rdatum_exists)
+		{
+			BOX			LRl = *union1,
+						LRr = *union2;
+			BOX			RLl = *union2,
+						RLr = *union1;
+			double		sizeLR,
+						sizeRL;
+
+			adjustBox(&LRl, DatumGetBoxP(v->spl_ldatum));
+			adjustBox(&LRr, DatumGetBoxP(v->spl_rdatum));
+			adjustBox(&RLl, DatumGetBoxP(v->spl_ldatum));
+			adjustBox(&RLr, DatumGetBoxP(v->spl_rdatum));
+
+			sizeLR = size_box(DirectFunctionCall2(rt_box_inter, BoxPGetDatum(&LRl), BoxPGetDatum(&LRr)));
+			sizeRL = size_box(DirectFunctionCall2(rt_box_inter, BoxPGetDatum(&RLl), BoxPGetDatum(&RLr)));
+
+			if (sizeLR > sizeRL)
+				firstToLeft = false;
+
+		}
+		else
+		{
+			float		p1,
+						p2;
+			GISTENTRY	oldUnion,
+						addon;
+
+			gistentryinit(oldUnion, (v->spl_ldatum_exists) ? v->spl_ldatum : v->spl_rdatum,
+						  NULL, NULL, InvalidOffsetNumber, FALSE);
+
+			gistentryinit(addon, BoxPGetDatum(union1), NULL, NULL, InvalidOffsetNumber, FALSE);
+			DirectFunctionCall3(gist_box_penalty, PointerGetDatum(&oldUnion), PointerGetDatum(&addon), PointerGetDatum(&p1));
+			gistentryinit(addon, BoxPGetDatum(union2), NULL, NULL, InvalidOffsetNumber, FALSE);
+			DirectFunctionCall3(gist_box_penalty, PointerGetDatum(&oldUnion), PointerGetDatum(&addon), PointerGetDatum(&p2));
+
+			if ((v->spl_ldatum_exists && p1 > p2) || (v->spl_rdatum_exists && p1 < p2))
+				firstToLeft = false;
+		}
+	}
+
+	if (firstToLeft)
+	{
+		v->spl_left = list1;
+		v->spl_right = list2;
+		v->spl_nleft = nlist1;
+		v->spl_nright = nlist2;
+		if (v->spl_ldatum_exists)
+			adjustBox(union1, DatumGetBoxP(v->spl_ldatum));
+		v->spl_ldatum = BoxPGetDatum(union1);
+		if (v->spl_rdatum_exists)
+			adjustBox(union2, DatumGetBoxP(v->spl_rdatum));
+		v->spl_rdatum = BoxPGetDatum(union2);
+	}
+	else
+	{
+		v->spl_left = list2;
+		v->spl_right = list1;
+		v->spl_nleft = nlist2;
+		v->spl_nright = nlist1;
+		if (v->spl_ldatum_exists)
+			adjustBox(union2, DatumGetBoxP(v->spl_ldatum));
+		v->spl_ldatum = BoxPGetDatum(union2);
+		if (v->spl_rdatum_exists)
+			adjustBox(union1, DatumGetBoxP(v->spl_rdatum));
+		v->spl_rdatum = BoxPGetDatum(union1);
+	}
+
+	v->spl_ldatum_exists = v->spl_rdatum_exists = false;
 }
 
 /*
@@ -162,7 +274,7 @@ adjustBox(BOX *b, BOX *addon)
 static void
 fallbackSplit(GistEntryVector *entryvec, GIST_SPLITVEC *v)
 {
-	OffsetNumber    i, 
+	OffsetNumber	i, 
 					maxoff;
 	BOX			   *unionL = NULL,
 				   *unionR = NULL;
@@ -202,9 +314,20 @@ fallbackSplit(GistEntryVector *entryvec, GIST_SPLITVEC *v)
 			}
 			else
 				adjustBox(unionR, cur);
+
 			v->spl_nright++;
 		}
 	}
+
+	if (v->spl_ldatum_exists)
+		adjustBox(unionL, DatumGetBoxP(v->spl_ldatum));
+	v->spl_ldatum = BoxPGetDatum(unionL);
+
+	if (v->spl_rdatum_exists)
+		adjustBox(unionR, DatumGetBoxP(v->spl_rdatum));
+	v->spl_rdatum = BoxPGetDatum(unionR);
+
+	v->spl_ldatum_exists = v->spl_rdatum_exists = false;
 }
 
 /*
@@ -389,23 +512,13 @@ gist_box_picksplit(PG_FUNCTION_ARGS)
 	}
 
 	if (direction == 'x')
-	{
-		v->spl_left = listL;
-		v->spl_right = listR;
-		v->spl_nleft = posL;
-		v->spl_nright = posR;
-		v->spl_ldatum = BoxPGetDatum(unionL);
-		v->spl_rdatum = BoxPGetDatum(unionR);
-	}
+		chooseLR(v,
+				 listL, posL, unionL,
+				 listR, posR, unionR);
 	else
-	{
-		v->spl_left = listB;
-		v->spl_right = listT;
-		v->spl_nleft = posB;
-		v->spl_nright = posT;
-		v->spl_ldatum = BoxPGetDatum(unionB);
-		v->spl_rdatum = BoxPGetDatum(unionT);
-	}
+		chooseLR(v,
+				 listB, posB, unionB,
+				 listT, posT, unionT);
 
 	PG_RETURN_POINTER(v);
 }
@@ -470,11 +583,13 @@ gist_box_leaf_consistent(BOX *key, BOX *query, StrategyNumber strategy)
 													PointerGetDatum(query)));
 			break;
 		case RTContainsStrategyNumber:
+		case RTOldContainsStrategyNumber:
 			retval = DatumGetBool(DirectFunctionCall2(box_contain,
 													  PointerGetDatum(key),
 													PointerGetDatum(query)));
 			break;
 		case RTContainedByStrategyNumber:
+		case RTOldContainedByStrategyNumber:
 			retval = DatumGetBool(DirectFunctionCall2(box_contained,
 													  PointerGetDatum(key),
 													PointerGetDatum(query)));
@@ -524,8 +639,6 @@ size_box(Datum dbox)
  *
  * We can use the same function since all types use bounding boxes as the
  * internal-page representation.
- *
- * This implements the same logic as the rtree internal-page strategy map.
  */
 static bool
 rtree_internal_consistent(BOX *key, BOX *query, StrategyNumber strategy)
@@ -561,11 +674,13 @@ rtree_internal_consistent(BOX *key, BOX *query, StrategyNumber strategy)
 			break;
 		case RTSameStrategyNumber:
 		case RTContainsStrategyNumber:
+		case RTOldContainsStrategyNumber:
 			retval = DatumGetBool(DirectFunctionCall2(box_contain,
 													  PointerGetDatum(key),
 													PointerGetDatum(query)));
 			break;
 		case RTContainedByStrategyNumber:
+		case RTOldContainedByStrategyNumber:
 			retval = DatumGetBool(DirectFunctionCall2(box_overlap,
 													  PointerGetDatum(key),
 													PointerGetDatum(query)));
@@ -621,14 +736,14 @@ gist_poly_compress(PG_FUNCTION_ARGS)
 			memcpy((void *) r, (void *) &(in->boundbox), sizeof(BOX));
 			gistentryinit(*retval, PointerGetDatum(r),
 						  entry->rel, entry->page,
-						  entry->offset, sizeof(BOX), FALSE);
+						  entry->offset, FALSE);
 
 		}
 		else
 		{
 			gistentryinit(*retval, (Datum) 0,
 						  entry->rel, entry->page,
-						  entry->offset, 0, FALSE);
+						  entry->offset, FALSE);
 		}
 	}
 	else
@@ -692,14 +807,14 @@ gist_circle_compress(PG_FUNCTION_ARGS)
 			r->low.y = in->center.y - in->radius;
 			gistentryinit(*retval, PointerGetDatum(r),
 						  entry->rel, entry->page,
-						  entry->offset, sizeof(BOX), FALSE);
+						  entry->offset, FALSE);
 
 		}
 		else
 		{
 			gistentryinit(*retval, (Datum) 0,
 						  entry->rel, entry->page,
-						  entry->offset, 0, FALSE);
+						  entry->offset, FALSE);
 		}
 	}
 	else
