@@ -3,11 +3,11 @@
  * twophase.c
  *		Two-phase commit support functions.
  *
- * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2006, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
- *		$PostgreSQL: pgsql/src/backend/access/transam/twophase.c,v 1.16.2.2 2007/02/13 19:39:55 tgl Exp $
+ *		$PostgreSQL: pgsql/src/backend/access/transam/twophase.c,v 1.25.2.3 2009/11/23 09:59:11 heikki Exp $
  *
  * NOTES
  *		Each global transaction is associated with a global transaction
@@ -44,6 +44,7 @@
 
 #include "access/heapam.h"
 #include "access/subtrans.h"
+#include "access/transam.h"
 #include "access/twophase.h"
 #include "access/twophase_rmgr.h"
 #include "access/xact.h"
@@ -52,10 +53,10 @@
 #include "miscadmin.h"
 #include "pgstat.h"
 #include "storage/fd.h"
-#include "storage/proc.h"
 #include "storage/procarray.h"
 #include "storage/smgr.h"
 #include "utils/builtins.h"
+#include "utils/memutils.h"
 
 
 /*
@@ -105,6 +106,7 @@ int			max_prepared_xacts = 5;
 typedef struct GlobalTransactionData
 {
 	PGPROC		proc;			/* dummy proc */
+	BackendId	dummyBackendId;	/* similar to backend id for backends */
 	TimestampTz prepared_at;	/* time of preparation */
 	XLogRecPtr	prepare_lsn;	/* XLOG offset of prepare record */
 	Oid			owner;			/* ID of user that executed the xact */
@@ -196,6 +198,20 @@ TwoPhaseShmemInit(void)
 		{
 			gxacts[i].proc.links.next = TwoPhaseState->freeGXacts;
 			TwoPhaseState->freeGXacts = MAKE_OFFSET(&gxacts[i]);
+
+			/*
+			 * Assign a unique ID for each dummy proc, so that the range of
+			 * dummy backend IDs immediately follows the range of normal
+			 * backend IDs. We don't dare to assign a real backend ID to
+			 * dummy procs, because prepared transactions don't take part in
+			 * cache invalidation like a real backend ID would imply, but
+			 * having a unique ID for them is nevertheless handy. This
+			 * arrangement allows you to allocate an array of size
+			 * (MaxBackends + max_prepared_xacts + 1), and have a slot for
+			 * every backend and prepared transaction. Currently multixact.c
+			 * uses that technique.
+			 */
+			gxacts[i].dummyBackendId = MaxBackends + 1 + i;
 		}
 	}
 	else
@@ -279,12 +295,14 @@ MarkAsPreparing(TransactionId xid, const char *gid,
 	gxact->proc.pid = 0;
 	gxact->proc.databaseId = databaseid;
 	gxact->proc.roleId = owner;
+	gxact->proc.inVacuum = false;
 	gxact->proc.lwWaiting = false;
 	gxact->proc.lwExclusive = false;
 	gxact->proc.lwWaitLink = NULL;
 	gxact->proc.waitLock = NULL;
 	gxact->proc.waitProcLock = NULL;
-	SHMQueueInit(&(gxact->proc.procLocks));
+	for (i = 0; i < NUM_LOCK_PARTITIONS; i++)
+		SHMQueueInit(&(gxact->proc.myProcLocks[i]));
 	/* subxid data must be filled later by GXactLoadSubxactData */
 	gxact->proc.subxids.overflowed = false;
 	gxact->proc.subxids.nxids = 0;
@@ -630,6 +648,22 @@ pg_prepared_xact(PG_FUNCTION_ARGS)
 	SRF_RETURN_DONE(funcctx);
 }
 
+/*
+ * TwoPhaseGetDummyProc
+ *		Get the dummy backend ID for prepared transaction specified by XID
+ *
+ * Dummy backend IDs are similar to real backend IDs of real backends.
+ * They start at MaxBackends + 1, and are unique across all currently active
+ * real backends and prepared transactions.
+ */
+BackendId
+TwoPhaseGetDummyBackendId(TransactionId xid)
+{
+	PGPROC *proc = TwoPhaseGetDummyProc(xid);
+
+	return ((GlobalTransaction) proc)->dummyBackendId;
+}
+
 /*
  * TwoPhaseGetDummyProc
  *		Get the PGPROC that represents a prepared transaction specified by XID
@@ -860,6 +894,15 @@ EndPrepare(GlobalTransaction gxact)
 	Assert(hdr->magic == TWOPHASE_MAGIC);
 	hdr->total_len = records.total_len + sizeof(pg_crc32);
 
+	/*
+	 * If the file size exceeds MaxAllocSize, we won't be able to read it in
+	 * ReadTwoPhaseFile. Check for that now, rather than fail at commit time.
+	 */
+	if (hdr->total_len > MaxAllocSize)
+		ereport(ERROR,
+				(errcode(ERRCODE_PROGRAM_LIMIT_EXCEEDED),
+				 errmsg("two-phase state file maximum length exceeded")));
+
 	/*
 	 * Create the 2PC state file.
 	 *
@@ -876,7 +919,7 @@ EndPrepare(GlobalTransaction gxact)
 	if (fd < 0)
 		ereport(ERROR,
 				(errcode_for_file_access(),
-				 errmsg("could not create twophase state file \"%s\": %m",
+				 errmsg("could not create two-phase state file \"%s\": %m",
 						path)));
 
 	/* Write data to file, and calculate CRC as we pass over it */
@@ -890,7 +933,7 @@ EndPrepare(GlobalTransaction gxact)
 			close(fd);
 			ereport(ERROR,
 					(errcode_for_file_access(),
-					 errmsg("could not write twophase state file: %m")));
+					 errmsg("could not write two-phase state file: %m")));
 		}
 	}
 
@@ -907,7 +950,7 @@ EndPrepare(GlobalTransaction gxact)
 		close(fd);
 		ereport(ERROR,
 				(errcode_for_file_access(),
-				 errmsg("could not write twophase state file: %m")));
+				 errmsg("could not write two-phase state file: %m")));
 	}
 
 	/* Back up to prepare for rewriting the CRC */
@@ -916,7 +959,7 @@ EndPrepare(GlobalTransaction gxact)
 		close(fd);
 		ereport(ERROR,
 				(errcode_for_file_access(),
-				 errmsg("could not seek in twophase state file: %m")));
+				 errmsg("could not seek in two-phase state file: %m")));
 	}
 
 	/*
@@ -956,13 +999,13 @@ EndPrepare(GlobalTransaction gxact)
 		close(fd);
 		ereport(ERROR,
 				(errcode_for_file_access(),
-				 errmsg("could not write twophase state file: %m")));
+				 errmsg("could not write two-phase state file: %m")));
 	}
 
 	if (close(fd) != 0)
 		ereport(ERROR,
 				(errcode_for_file_access(),
-				 errmsg("could not close twophase state file: %m")));
+				 errmsg("could not close two-phase state file: %m")));
 
 	/*
 	 * Mark the prepared transaction as valid.	As soon as xact.c marks MyProc
@@ -1032,21 +1075,23 @@ ReadTwoPhaseFile(TransactionId xid)
 	{
 		ereport(WARNING,
 				(errcode_for_file_access(),
-				 errmsg("could not open twophase state file \"%s\": %m",
+				 errmsg("could not open two-phase state file \"%s\": %m",
 						path)));
 		return NULL;
 	}
 
 	/*
 	 * Check file length.  We can determine a lower bound pretty easily. We
-	 * set an upper bound mainly to avoid palloc() failure on a corrupt file.
+	 * set an upper bound to avoid palloc() failure on a corrupt file, though
+	 * we can't guarantee that we won't get an out of memory error anyway,
+	 * even on a valid file.
 	 */
 	if (fstat(fd, &stat))
 	{
 		close(fd);
 		ereport(WARNING,
 				(errcode_for_file_access(),
-				 errmsg("could not stat twophase state file \"%s\": %m",
+				 errmsg("could not stat two-phase state file \"%s\": %m",
 						path)));
 		return NULL;
 	}
@@ -1054,7 +1099,7 @@ ReadTwoPhaseFile(TransactionId xid)
 	if (stat.st_size < (MAXALIGN(sizeof(TwoPhaseFileHeader)) +
 						MAXALIGN(sizeof(TwoPhaseRecordOnDisk)) +
 						sizeof(pg_crc32)) ||
-		stat.st_size > 10000000)
+		stat.st_size > MaxAllocSize)
 	{
 		close(fd);
 		return NULL;
@@ -1077,7 +1122,7 @@ ReadTwoPhaseFile(TransactionId xid)
 		close(fd);
 		ereport(WARNING,
 				(errcode_for_file_access(),
-				 errmsg("could not read twophase state file \"%s\": %m",
+				 errmsg("could not read two-phase state file \"%s\": %m",
 						path)));
 		pfree(buf);
 		return NULL;
@@ -1138,7 +1183,7 @@ FinishPreparedTransaction(const char *gid, bool isCommit)
 	if (buf == NULL)
 		ereport(ERROR,
 				(errcode(ERRCODE_DATA_CORRUPTED),
-				 errmsg("twophase state file for transaction %u is corrupt",
+				 errmsg("two-phase state file for transaction %u is corrupt",
 						xid)));
 
 	/*
@@ -1260,8 +1305,8 @@ RemoveTwoPhaseFile(TransactionId xid, bool giveWarning)
 		if (errno != ENOENT || giveWarning)
 			ereport(WARNING,
 					(errcode_for_file_access(),
-				   errmsg("could not remove two-phase state file \"%s\": %m",
-						  path)));
+					 errmsg("could not remove two-phase state file \"%s\": %m",
+							path)));
 }
 
 /*
@@ -1289,7 +1334,7 @@ RecreateTwoPhaseFile(TransactionId xid, void *content, int len)
 	if (fd < 0)
 		ereport(ERROR,
 				(errcode_for_file_access(),
-				 errmsg("could not recreate twophase state file \"%s\": %m",
+				 errmsg("could not recreate two-phase state file \"%s\": %m",
 						path)));
 
 	/* Write content and CRC */
@@ -1298,14 +1343,14 @@ RecreateTwoPhaseFile(TransactionId xid, void *content, int len)
 		close(fd);
 		ereport(ERROR,
 				(errcode_for_file_access(),
-				 errmsg("could not write twophase state file: %m")));
+				 errmsg("could not write two-phase state file: %m")));
 	}
 	if (write(fd, &statefile_crc, sizeof(pg_crc32)) != sizeof(pg_crc32))
 	{
 		close(fd);
 		ereport(ERROR,
 				(errcode_for_file_access(),
-				 errmsg("could not write twophase state file: %m")));
+				 errmsg("could not write two-phase state file: %m")));
 	}
 
 	/*
@@ -1317,13 +1362,13 @@ RecreateTwoPhaseFile(TransactionId xid, void *content, int len)
 		close(fd);
 		ereport(ERROR,
 				(errcode_for_file_access(),
-				 errmsg("could not fsync twophase state file: %m")));
+				 errmsg("could not fsync two-phase state file: %m")));
 	}
 
 	if (close(fd) != 0)
 		ereport(ERROR,
 				(errcode_for_file_access(),
-				 errmsg("could not close twophase state file: %m")));
+				 errmsg("could not close two-phase state file: %m")));
 }
 
 /*
@@ -1400,7 +1445,7 @@ CheckPointTwoPhase(XLogRecPtr redo_horizon)
 			}
 			ereport(ERROR,
 					(errcode_for_file_access(),
-					 errmsg("could not open twophase state file \"%s\": %m",
+					 errmsg("could not open two-phase state file \"%s\": %m",
 							path)));
 		}
 
@@ -1409,14 +1454,14 @@ CheckPointTwoPhase(XLogRecPtr redo_horizon)
 			close(fd);
 			ereport(ERROR,
 					(errcode_for_file_access(),
-					 errmsg("could not fsync twophase state file \"%s\": %m",
+					 errmsg("could not fsync two-phase state file \"%s\": %m",
 							path)));
 		}
 
 		if (close(fd) != 0)
 			ereport(ERROR,
 					(errcode_for_file_access(),
-					 errmsg("could not close twophase state file \"%s\": %m",
+					 errmsg("could not close two-phase state file \"%s\": %m",
 							path)));
 	}
 
@@ -1472,7 +1517,7 @@ PrescanPreparedTransactions(void)
 			if (TransactionIdFollowsOrEquals(xid, origNextXid))
 			{
 				ereport(WARNING,
-						(errmsg("removing future twophase state file \"%s\"",
+						(errmsg("removing future two-phase state file \"%s\"",
 								clde->d_name)));
 				RemoveTwoPhaseFile(xid, true);
 				continue;
@@ -1488,7 +1533,7 @@ PrescanPreparedTransactions(void)
 			if (buf == NULL)
 			{
 				ereport(WARNING,
-						(errmsg("removing corrupt twophase state file \"%s\"",
+						(errmsg("removing corrupt two-phase state file \"%s\"",
 								clde->d_name)));
 				RemoveTwoPhaseFile(xid, true);
 				continue;
@@ -1499,7 +1544,7 @@ PrescanPreparedTransactions(void)
 			if (!TransactionIdEquals(hdr->xid, xid))
 			{
 				ereport(WARNING,
-						(errmsg("removing corrupt twophase state file \"%s\"",
+						(errmsg("removing corrupt two-phase state file \"%s\"",
 								clde->d_name)));
 				RemoveTwoPhaseFile(xid, true);
 				pfree(buf);
@@ -1576,7 +1621,7 @@ RecoverPreparedTransactions(void)
 			if (TransactionIdDidCommit(xid) || TransactionIdDidAbort(xid))
 			{
 				ereport(WARNING,
-						(errmsg("removing stale twophase state file \"%s\"",
+						(errmsg("removing stale two-phase state file \"%s\"",
 								clde->d_name)));
 				RemoveTwoPhaseFile(xid, true);
 				continue;
@@ -1587,7 +1632,7 @@ RecoverPreparedTransactions(void)
 			if (buf == NULL)
 			{
 				ereport(WARNING,
-						(errmsg("removing corrupt twophase state file \"%s\"",
+						(errmsg("removing corrupt two-phase state file \"%s\"",
 								clde->d_name)));
 				RemoveTwoPhaseFile(xid, true);
 				continue;
