@@ -3,18 +3,17 @@
  * planagg.c
  *	  Special planning for aggregate queries.
  *
- * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2006, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/optimizer/plan/planagg.c,v 1.10.2.2 2006/04/28 20:57:59 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/optimizer/plan/planagg.c,v 1.22.2.1 2007/02/06 06:50:33 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
 #include "postgres.h"
 
-#include "access/skey.h"
 #include "catalog/pg_aggregate.h"
 #include "catalog/pg_type.h"
 #include "nodes/makefuncs.h"
@@ -24,9 +23,9 @@
 #include "optimizer/paths.h"
 #include "optimizer/planmain.h"
 #include "optimizer/subselect.h"
-#include "parser/parsetree.h"
 #include "parser/parse_clause.h"
 #include "parser/parse_expr.h"
+#include "parser/parsetree.h"
 #include "utils/lsyscache.h"
 #include "utils/syscache.h"
 
@@ -46,8 +45,7 @@ static bool build_minmax_path(PlannerInfo *root, RelOptInfo *rel,
 				  MinMaxAggInfo *info);
 static ScanDirection match_agg_to_index_col(MinMaxAggInfo *info,
 					   IndexOptInfo *index, int indexcol);
-static void make_agg_subplan(PlannerInfo *root, MinMaxAggInfo *info,
-				 List *constant_quals);
+static void make_agg_subplan(PlannerInfo *root, MinMaxAggInfo *info);
 static Node *replace_aggs_with_params_mutator(Node *node, List **context);
 static Oid	fetch_agg_sort_op(Oid aggfnoid);
 
@@ -71,6 +69,7 @@ Plan *
 optimize_minmax_aggregates(PlannerInfo *root, List *tlist, Path *best_path)
 {
 	Query	   *parse = root->parse;
+	FromExpr   *jtnode;
 	RangeTblRef *rtr;
 	RangeTblEntry *rte;
 	RelOptInfo *rel;
@@ -81,7 +80,6 @@ optimize_minmax_aggregates(PlannerInfo *root, List *tlist, Path *best_path)
 	Plan	   *plan;
 	Node	   *hqual;
 	QualCost	tlist_cost;
-	List	   *constant_quals;
 
 	/* Nothing to do if query has no aggregates */
 	if (!parse->hasAggs)
@@ -104,14 +102,19 @@ optimize_minmax_aggregates(PlannerInfo *root, List *tlist, Path *best_path)
 	 * We also restrict the query to reference exactly one table, since join
 	 * conditions can't be handled reasonably.  (We could perhaps handle a
 	 * query containing cartesian-product joins, but it hardly seems worth the
-	 * trouble.)
+	 * trouble.)  However, the single real table could be buried in several
+	 * levels of FromExpr.
 	 */
-	Assert(parse->jointree != NULL && IsA(parse->jointree, FromExpr));
-	if (list_length(parse->jointree->fromlist) != 1)
-		return NULL;
-	rtr = (RangeTblRef *) linitial(parse->jointree->fromlist);
-	if (!IsA(rtr, RangeTblRef))
+	jtnode = parse->jointree;
+	while (IsA(jtnode, FromExpr))
+	{
+		if (list_length(jtnode->fromlist) != 1)
+			return NULL;
+		jtnode = linitial(jtnode->fromlist);
+	}
+	if (!IsA(jtnode, RangeTblRef))
 		return NULL;
+	rtr = (RangeTblRef *) jtnode;
 	rte = rt_fetch(rtr->rtindex, parse->rtable);
 	if (rte->rtekind != RTE_RELATION || rte->inh)
 		return NULL;
@@ -164,27 +167,13 @@ optimize_minmax_aggregates(PlannerInfo *root, List *tlist, Path *best_path)
 		return NULL;			/* too expensive */
 
 	/*
-	 * OK, we are going to generate an optimized plan.	The first thing we
-	 * need to do is look for any non-variable WHERE clauses that
-	 * query_planner might have removed from the basic plan.  (Normal WHERE
-	 * clauses will be properly incorporated into the sub-plans by
-	 * create_plan.)  If there are any, they will be in a gating Result node
-	 * atop the best_path. They have to be incorporated into a gating Result
-	 * in each sub-plan in order to produce the semantically correct result.
+	 * OK, we are going to generate an optimized plan.
 	 */
-	if (IsA(best_path, ResultPath))
-	{
-		constant_quals = ((ResultPath *) best_path)->constantqual;
-		/* no need to do this more than once: */
-		constant_quals = order_qual_clauses(root, constant_quals);
-	}
-	else
-		constant_quals = NIL;
 
 	/* Pass 3: generate subplans and output Param nodes */
 	foreach(l, aggs_list)
 	{
-		make_agg_subplan(root, (MinMaxAggInfo *) lfirst(l), constant_quals);
+		make_agg_subplan(root, (MinMaxAggInfo *) lfirst(l));
 	}
 
 	/*
@@ -234,12 +223,13 @@ find_minmax_aggs_walker(Node *node, List **context)
 	{
 		Aggref	   *aggref = (Aggref *) node;
 		Oid			aggsortop;
+		Expr	   *curTarget;
 		MinMaxAggInfo *info;
 		ListCell   *l;
 
 		Assert(aggref->agglevelsup == 0);
-		if (aggref->aggstar)
-			return true;		/* foo(*) is surely not optimizable */
+		if (list_length(aggref->args) != 1)
+			return true;		/* it couldn't be MIN/MAX */
 		/* note: we do not care if DISTINCT is mentioned ... */
 
 		aggsortop = fetch_agg_sort_op(aggref->aggfnoid);
@@ -249,18 +239,19 @@ find_minmax_aggs_walker(Node *node, List **context)
 		/*
 		 * Check whether it's already in the list, and add it if not.
 		 */
+		curTarget = linitial(aggref->args);
 		foreach(l, *context)
 		{
 			info = (MinMaxAggInfo *) lfirst(l);
 			if (info->aggfnoid == aggref->aggfnoid &&
-				equal(info->target, aggref->target))
+				equal(info->target, curTarget))
 				return false;
 		}
 
 		info = (MinMaxAggInfo *) palloc0(sizeof(MinMaxAggInfo));
 		info->aggfnoid = aggref->aggfnoid;
 		info->aggsortop = aggsortop;
-		info->target = aggref->target;
+		info->target = curTarget;
 
 		*context = lappend(*context, info);
 
@@ -338,6 +329,7 @@ build_minmax_path(PlannerInfo *root, RelOptInfo *rel, MinMaxAggInfo *info)
 												index->rel->baserestrictinfo,
 													NIL,
 													NULL,
+													SAOP_FORBID,
 													&found_clause);
 
 		if (list_length(restrictclauses) < indexcol)
@@ -372,7 +364,7 @@ build_minmax_path(PlannerInfo *root, RelOptInfo *rel, MinMaxAggInfo *info)
 									 restrictclauses,
 									 NIL,
 									 indexscandir,
-									 false);
+									 NULL);
 
 		/*
 		 * Estimate actual cost of fetching just one row.
@@ -433,11 +425,12 @@ match_agg_to_index_col(MinMaxAggInfo *info, IndexOptInfo *index, int indexcol)
  * Construct a suitable plan for a converted aggregate query
  */
 static void
-make_agg_subplan(PlannerInfo *root, MinMaxAggInfo *info, List *constant_quals)
+make_agg_subplan(PlannerInfo *root, MinMaxAggInfo *info)
 {
 	PlannerInfo subroot;
 	Query	   *subparse;
 	Plan	   *plan;
+	Plan	   *iplan;
 	TargetEntry *tle;
 	SortClause *sortcl;
 	NullTest   *ntest;
@@ -452,6 +445,7 @@ make_agg_subplan(PlannerInfo *root, MinMaxAggInfo *info, List *constant_quals)
 	subparse->commandType = CMD_SELECT;
 	subparse->resultRelation = 0;
 	subparse->resultRelations = NIL;
+	subparse->returningLists = NIL;
 	subparse->into = NULL;
 	subparse->hasAggs = false;
 	subparse->groupClause = NIL;
@@ -474,15 +468,14 @@ make_agg_subplan(PlannerInfo *root, MinMaxAggInfo *info, List *constant_quals)
 
 	/* set up LIMIT 1 */
 	subparse->limitOffset = NULL;
-	subparse->limitCount = (Node *) makeConst(INT4OID, sizeof(int4),
-											  Int32GetDatum(1),
-											  false, true);
+	subparse->limitCount = (Node *) makeConst(INT8OID, sizeof(int64),
+											  Int64GetDatum(1),
+											  false, false /* not by val */ );
 
 	/*
 	 * Generate the plan for the subquery.	We already have a Path for the
 	 * basic indexscan, but we have to convert it to a Plan and attach a LIMIT
-	 * node above it.  We might need a gating Result, too, to handle any
-	 * non-variable qual clauses.
+	 * node above it.
 	 *
 	 * Also we must add a "WHERE foo IS NOT NULL" restriction to the
 	 * indexscan, to be sure we don't return a NULL, which'd be contrary to
@@ -490,21 +483,26 @@ make_agg_subplan(PlannerInfo *root, MinMaxAggInfo *info, List *constant_quals)
 	 * earlier, so that the selectivity of the restriction could be included
 	 * in our cost estimates.  But that looks painful, and in most cases the
 	 * fraction of NULLs isn't high enough to change the decision.
+	 *
+	 * The NOT NULL qual has to go on the actual indexscan; create_plan might
+	 * have stuck a gating Result atop that, if there were any pseudoconstant
+	 * quals.
 	 */
 	plan = create_plan(&subroot, (Path *) info->path);
 
 	plan->targetlist = copyObject(subparse->targetList);
 
+	if (IsA(plan, Result))
+		iplan = plan->lefttree;
+	else
+		iplan = plan;
+	Assert(IsA(iplan, IndexScan));
+
 	ntest = makeNode(NullTest);
 	ntest->nulltesttype = IS_NOT_NULL;
 	ntest->arg = copyObject(info->target);
 
-	plan->qual = lcons(ntest, plan->qual);
-
-	if (constant_quals)
-		plan = (Plan *) make_result(copyObject(plan->targetlist),
-									copyObject(constant_quals),
-									plan);
+	iplan->qual = lcons(ntest, iplan->qual);
 
 	plan = (Plan *) make_limit(plan,
 							   subparse->limitOffset,
@@ -531,13 +529,14 @@ replace_aggs_with_params_mutator(Node *node, List **context)
 	{
 		Aggref	   *aggref = (Aggref *) node;
 		ListCell   *l;
+		Expr	   *curTarget = linitial(aggref->args);
 
 		foreach(l, *context)
 		{
 			MinMaxAggInfo *info = (MinMaxAggInfo *) lfirst(l);
 
 			if (info->aggfnoid == aggref->aggfnoid &&
-				equal(info->target, aggref->target))
+				equal(info->target, curTarget))
 				return (Node *) info->param;
 		}
 		elog(ERROR, "failed to re-find aggregate info record");
