@@ -8,7 +8,7 @@
  *
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/parser/parse_expr.c,v 1.198.2.4 2010/07/30 17:57:12 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/parser/parse_expr.c,v 1.198.2.2 2008/10/26 02:46:36 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -587,15 +587,12 @@ transformAExprOp(ParseState *pstate, A_Expr *a)
 	/*
 	 * Special-case "foo = NULL" and "NULL = foo" for compatibility with
 	 * standards-broken products (like Microsoft's).  Turn these into IS NULL
-	 * exprs. (If either side is a CaseTestExpr, then the expression was
-	 * generated internally from a CASE-WHEN expression, and
-	 * transform_null_equals does not apply.)
+	 * exprs.
 	 */
 	if (Transform_null_equals &&
 		list_length(a->name) == 1 &&
 		strcmp(strVal(linitial(a->name)), "=") == 0 &&
-		(exprIsNullConstant(lexpr) || exprIsNullConstant(rexpr)) &&
-		(!IsA(lexpr, CaseTestExpr) && !IsA(rexpr, CaseTestExpr)))
+		(exprIsNullConstant(lexpr) || exprIsNullConstant(rexpr)))
 	{
 		NullTest   *n = makeNode(NullTest);
 
