@@ -7,12 +7,12 @@
  * detection and resolution algorithms.
  *
  *
- * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2006, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/storage/lmgr/deadlock.c,v 1.32 2004/12/31 22:01:05 pgsql Exp $
+ *	  $PostgreSQL: pgsql/src/backend/storage/lmgr/deadlock.c,v 1.43 2006/09/22 23:20:13 tgl Exp $
  *
  *	Interface:
  *
@@ -25,7 +25,6 @@
  */
 #include "postgres.h"
 
-#include "catalog/pg_class.h"
 #include "lib/stringinfo.h"
 #include "miscadmin.h"
 #include "storage/proc.h"
@@ -53,9 +52,9 @@ typedef struct
  * Information saved about each edge in a detected deadlock cycle.	This
  * is used to print a diagnostic message upon failure.
  *
- * Note: because we want to examine this info after releasing the LockMgrLock,
- * we can't just store LOCK and PGPROC pointers; we must extract out all the
- * info we want to be able to print.
+ * Note: because we want to examine this info after releasing the lock
+ * manager's partition locks, we can't just store LOCK and PGPROC pointers;
+ * we must extract out all the info we want to be able to print.
  */
 typedef struct
 {
@@ -130,15 +129,15 @@ InitDeadLockChecking(void)
 	oldcxt = MemoryContextSwitchTo(TopMemoryContext);
 
 	/*
-	 * FindLockCycle needs at most MaxBackends entries in visitedProcs[]
-	 * and deadlockDetails[].
+	 * FindLockCycle needs at most MaxBackends entries in visitedProcs[] and
+	 * deadlockDetails[].
 	 */
 	visitedProcs = (PGPROC **) palloc(MaxBackends * sizeof(PGPROC *));
 	deadlockDetails = (DEADLOCK_INFO *) palloc(MaxBackends * sizeof(DEADLOCK_INFO));
 
 	/*
-	 * TopoSort needs to consider at most MaxBackends wait-queue entries,
-	 * and it needn't run concurrently with FindLockCycle.
+	 * TopoSort needs to consider at most MaxBackends wait-queue entries, and
+	 * it needn't run concurrently with FindLockCycle.
 	 */
 	topoProcs = visitedProcs;	/* re-use this space */
 	beforeConstraints = (int *) palloc(MaxBackends * sizeof(int));
@@ -146,33 +145,32 @@ InitDeadLockChecking(void)
 
 	/*
 	 * We need to consider rearranging at most MaxBackends/2 wait queues
-	 * (since it takes at least two waiters in a queue to create a soft
-	 * edge), and the expanded form of the wait queues can't involve more
-	 * than MaxBackends total waiters.
+	 * (since it takes at least two waiters in a queue to create a soft edge),
+	 * and the expanded form of the wait queues can't involve more than
+	 * MaxBackends total waiters.
 	 */
 	waitOrders = (WAIT_ORDER *)
 		palloc((MaxBackends / 2) * sizeof(WAIT_ORDER));
 	waitOrderProcs = (PGPROC **) palloc(MaxBackends * sizeof(PGPROC *));
 
 	/*
-	 * Allow at most MaxBackends distinct constraints in a configuration.
-	 * (Is this enough?  In practice it seems it should be, but I don't
-	 * quite see how to prove it.  If we run out, we might fail to find a
-	 * workable wait queue rearrangement even though one exists.)  NOTE
-	 * that this number limits the maximum recursion depth of
-	 * DeadLockCheckRecurse. Making it really big might potentially allow
-	 * a stack-overflow problem.
+	 * Allow at most MaxBackends distinct constraints in a configuration. (Is
+	 * this enough?  In practice it seems it should be, but I don't quite see
+	 * how to prove it.  If we run out, we might fail to find a workable wait
+	 * queue rearrangement even though one exists.)  NOTE that this number
+	 * limits the maximum recursion depth of DeadLockCheckRecurse. Making it
+	 * really big might potentially allow a stack-overflow problem.
 	 */
 	maxCurConstraints = MaxBackends;
 	curConstraints = (EDGE *) palloc(maxCurConstraints * sizeof(EDGE));
 
 	/*
 	 * Allow up to 3*MaxBackends constraints to be saved without having to
-	 * re-run TestConfiguration.  (This is probably more than enough, but
-	 * we can survive if we run low on space by doing excess runs of
-	 * TestConfiguration to re-compute constraint lists each time needed.)
-	 * The last MaxBackends entries in possibleConstraints[] are reserved
-	 * as output workspace for FindLockCycle.
+	 * re-run TestConfiguration.  (This is probably more than enough, but we
+	 * can survive if we run low on space by doing excess runs of
+	 * TestConfiguration to re-compute constraint lists each time needed.) The
+	 * last MaxBackends entries in possibleConstraints[] are reserved as
+	 * output workspace for FindLockCycle.
 	 */
 	maxPossibleConstraints = MaxBackends * 4;
 	possibleConstraints =
@@ -189,19 +187,11 @@ InitDeadLockChecking(void)
  * deadlock.  If resolution is impossible, return TRUE --- the caller
  * is then expected to abort the given proc's transaction.
  *
- * We can't block on user locks, so no sense testing for deadlock
- * because there is no blocking, and no timer for the block.  So,
- * only look at regular locks.
- *
- * We must have already locked the master lock before being called.
- * NOTE: although the lockmethod structure appears to allow each lock
- * table to have a different masterLock, all locks that can block had
- * better use the same LWLock, else this code will not be adequately
- * interlocked!
+ * Caller must already have locked all partitions of the lock tables.
  *
  * On failure, deadlock details are recorded in deadlockDetails[] for
  * subsequent printing by DeadLockReport().  That activity is separate
- * because we don't want to do it while holding the master lock.
+ * because we don't want to do it while holding all those LWLocks.
  */
 bool
 DeadLockCheck(PGPROC *proc)
@@ -361,9 +351,9 @@ TestConfiguration(PGPROC *startProc)
 		return -1;
 
 	/*
-	 * Check for cycles involving startProc or any of the procs mentioned
-	 * in constraints.	We check startProc last because if it has a soft
-	 * cycle still to be dealt with, we want to deal with that first.
+	 * Check for cycles involving startProc or any of the procs mentioned in
+	 * constraints.  We check startProc last because if it has a soft cycle
+	 * still to be dealt with, we want to deal with that first.
 	 */
 	for (i = 0; i < nCurConstraints; i++)
 	{
@@ -447,8 +437,8 @@ FindLockCycleRecurse(PGPROC *checkProc,
 			if (i == 0)
 			{
 				/*
-				 * record total length of cycle --- outer levels will now
-				 * fill deadlockDetails[]
+				 * record total length of cycle --- outer levels will now fill
+				 * deadlockDetails[]
 				 */
 				Assert(depth <= MaxBackends);
 				nDeadlockDetails = depth;
@@ -457,8 +447,8 @@ FindLockCycleRecurse(PGPROC *checkProc,
 			}
 
 			/*
-			 * Otherwise, we have a cycle but it does not include the
-			 * start point, so say "no deadlock".
+			 * Otherwise, we have a cycle but it does not include the start
+			 * point, so say "no deadlock".
 			 */
 			return false;
 		}
@@ -480,8 +470,8 @@ FindLockCycleRecurse(PGPROC *checkProc,
 	conflictMask = lockMethodTable->conflictTab[checkProc->waitLockMode];
 
 	/*
-	 * Scan for procs that already hold conflicting locks.	These are
-	 * "hard" edges in the waits-for graph.
+	 * Scan for procs that already hold conflicting locks.	These are "hard"
+	 * edges in the waits-for graph.
 	 */
 	procLocks = &(lock->procLocks);
 
@@ -490,7 +480,7 @@ FindLockCycleRecurse(PGPROC *checkProc,
 
 	while (proclock)
 	{
-		proc = (PGPROC *) MAKE_PTR(proclock->tag.proc);
+		proc = proclock->tag.myProc;
 
 		/* A proc never blocks itself */
 		if (proc != checkProc)
@@ -520,15 +510,14 @@ FindLockCycleRecurse(PGPROC *checkProc,
 		}
 
 		proclock = (PROCLOCK *) SHMQueueNext(procLocks, &proclock->lockLink,
-										   offsetof(PROCLOCK, lockLink));
+											 offsetof(PROCLOCK, lockLink));
 	}
 
 	/*
 	 * Scan for procs that are ahead of this one in the lock's wait queue.
-	 * Those that have conflicting requests soft-block this one.  This
-	 * must be done after the hard-block search, since if another proc
-	 * both hard- and soft-blocks this one, we want to call it a hard
-	 * edge.
+	 * Those that have conflicting requests soft-block this one.  This must be
+	 * done after the hard-block search, since if another proc both hard- and
+	 * soft-blocks this one, we want to call it a hard edge.
 	 *
 	 * If there is a proposed re-ordering of the lock's wait order, use that
 	 * rather than the current wait order.
@@ -569,8 +558,7 @@ FindLockCycleRecurse(PGPROC *checkProc,
 					info->pid = checkProc->pid;
 
 					/*
-					 * Add this edge to the list of soft edges in the
-					 * cycle
+					 * Add this edge to the list of soft edges in the cycle
 					 */
 					Assert(*nSoftEdges < MaxBackends);
 					softEdges[*nSoftEdges].waiter = checkProc;
@@ -610,8 +598,7 @@ FindLockCycleRecurse(PGPROC *checkProc,
 					info->pid = checkProc->pid;
 
 					/*
-					 * Add this edge to the list of soft edges in the
-					 * cycle
+					 * Add this edge to the list of soft edges in the cycle
 					 */
 					Assert(*nSoftEdges < MaxBackends);
 					softEdges[*nSoftEdges].waiter = checkProc;
@@ -655,8 +642,8 @@ ExpandConstraints(EDGE *constraints,
 
 	/*
 	 * Scan constraint list backwards.	This is because the last-added
-	 * constraint is the only one that could fail, and so we want to test
-	 * it for inconsistency first.
+	 * constraint is the only one that could fail, and so we want to test it
+	 * for inconsistency first.
 	 */
 	for (i = nConstraints; --i >= 0;)
 	{
@@ -679,8 +666,8 @@ ExpandConstraints(EDGE *constraints,
 		Assert(nWaitOrderProcs <= MaxBackends);
 
 		/*
-		 * Do the topo sort.  TopoSort need not examine constraints after
-		 * this one, since they must be for different locks.
+		 * Do the topo sort.  TopoSort need not examine constraints after this
+		 * one, since they must be for different locks.
 		 */
 		if (!TopoSort(lock, constraints, i + 1,
 					  waitOrders[nWaitOrders].procs))
@@ -739,15 +726,14 @@ TopoSort(LOCK *lock,
 	}
 
 	/*
-	 * Scan the constraints, and for each proc in the array, generate a
-	 * count of the number of constraints that say it must be before
-	 * something else, plus a list of the constraints that say it must be
-	 * after something else. The count for the j'th proc is stored in
-	 * beforeConstraints[j], and the head of its list in
-	 * afterConstraints[j].  Each constraint stores its list link in
-	 * constraints[i].link (note any constraint will be in just one list).
-	 * The array index for the before-proc of the i'th constraint is
-	 * remembered in constraints[i].pred.
+	 * Scan the constraints, and for each proc in the array, generate a count
+	 * of the number of constraints that say it must be before something else,
+	 * plus a list of the constraints that say it must be after something
+	 * else. The count for the j'th proc is stored in beforeConstraints[j],
+	 * and the head of its list in afterConstraints[j].  Each constraint
+	 * stores its list link in constraints[i].link (note any constraint will
+	 * be in just one list). The array index for the before-proc of the i'th
+	 * constraint is remembered in constraints[i].pred.
 	 */
 	MemSet(beforeConstraints, 0, queue_size * sizeof(int));
 	MemSet(afterConstraints, 0, queue_size * sizeof(int));
@@ -836,6 +822,79 @@ PrintLockQueue(LOCK *lock, const char *info)
 }
 #endif
 
+/*
+ * Append a description of a lockable object to buf.
+ *
+ * XXX probably this should be exported from lmgr.c or some such place.
+ */
+static void
+DescribeLockTag(StringInfo buf, const LOCKTAG *lock)
+{
+	switch (lock->locktag_type)
+	{
+		case LOCKTAG_RELATION:
+			appendStringInfo(buf,
+							 _("relation %u of database %u"),
+							 lock->locktag_field2,
+							 lock->locktag_field1);
+			break;
+		case LOCKTAG_RELATION_EXTEND:
+			appendStringInfo(buf,
+							 _("extension of relation %u of database %u"),
+							 lock->locktag_field2,
+							 lock->locktag_field1);
+			break;
+		case LOCKTAG_PAGE:
+			appendStringInfo(buf,
+							 _("page %u of relation %u of database %u"),
+							 lock->locktag_field3,
+							 lock->locktag_field2,
+							 lock->locktag_field1);
+			break;
+		case LOCKTAG_TUPLE:
+			appendStringInfo(buf,
+							 _("tuple (%u,%u) of relation %u of database %u"),
+							 lock->locktag_field3,
+							 lock->locktag_field4,
+							 lock->locktag_field2,
+							 lock->locktag_field1);
+			break;
+		case LOCKTAG_TRANSACTION:
+			appendStringInfo(buf,
+							 _("transaction %u"),
+							 lock->locktag_field1);
+			break;
+		case LOCKTAG_OBJECT:
+			appendStringInfo(buf,
+							 _("object %u of class %u of database %u"),
+							 lock->locktag_field3,
+							 lock->locktag_field2,
+							 lock->locktag_field1);
+			break;
+		case LOCKTAG_USERLOCK:
+			/* reserved for old contrib code, now on pgfoundry */
+			appendStringInfo(buf,
+							 _("user lock [%u,%u,%u]"),
+							 lock->locktag_field1,
+							 lock->locktag_field2,
+							 lock->locktag_field3);
+			break;
+		case LOCKTAG_ADVISORY:
+			appendStringInfo(buf,
+							 _("advisory lock [%u,%u,%u,%u]"),
+							 lock->locktag_field1,
+							 lock->locktag_field2,
+							 lock->locktag_field3,
+							 lock->locktag_field4);
+			break;
+		default:
+			appendStringInfo(buf,
+							 _("unrecognized locktag type %d"),
+							 lock->locktag_type);
+			break;
+	}
+}
+
 /*
  * Report a detected deadlock, with available details.
  */
@@ -843,9 +902,12 @@ void
 DeadLockReport(void)
 {
 	StringInfoData buf;
+	StringInfoData buf2;
 	int			i;
 
 	initStringInfo(&buf);
+	initStringInfo(&buf2);
+
 	for (i = 0; i < nDeadlockDetails; i++)
 	{
 		DEADLOCK_INFO *info = &deadlockDetails[i];
@@ -860,27 +922,19 @@ DeadLockReport(void)
 		if (i > 0)
 			appendStringInfoChar(&buf, '\n');
 
-		if (info->locktag.relId == XactLockTableId && info->locktag.dbId == 0)
-		{
-			/* Lock is for transaction ID */
-			appendStringInfo(&buf,
-							 gettext("Process %d waits for %s on transaction %u; blocked by process %d."),
-							 info->pid,
-							 GetLockmodeName(info->lockmode),
-							 info->locktag.objId.xid,
-							 nextpid);
-		}
-		else
-		{
-			/* Lock is for a relation */
-			appendStringInfo(&buf,
-							 gettext("Process %d waits for %s on relation %u of database %u; blocked by process %d."),
-							 info->pid,
-							 GetLockmodeName(info->lockmode),
-							 info->locktag.relId,
-							 info->locktag.dbId,
-							 nextpid);
-		}
+		/* reset buf2 to hold next object description */
+		buf2.len = 0;
+		buf2.data[0] = '\0';
+
+		DescribeLockTag(&buf2, &info->locktag);
+
+		appendStringInfo(&buf,
+				  _("Process %d waits for %s on %s; blocked by process %d."),
+						 info->pid,
+						 GetLockmodeName(info->locktag.locktag_lockmethodid,
+										 info->lockmode),
+						 buf2.data,
+						 nextpid);
 	}
 	ereport(ERROR,
 			(errcode(ERRCODE_T_R_DEADLOCK_DETECTED),
