@@ -8,12 +8,12 @@
  * special I/O conversion routines.
  *
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2006, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/backend/utils/adt/regproc.c,v 1.82 2003/09/25 06:58:04 petere Exp $
+ *	  $PostgreSQL: pgsql/src/backend/utils/adt/regproc.c,v 1.99 2006/07/14 14:52:24 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -23,13 +23,11 @@
 
 #include "access/genam.h"
 #include "access/heapam.h"
-#include "catalog/catname.h"
 #include "catalog/indexing.h"
 #include "catalog/namespace.h"
 #include "catalog/pg_operator.h"
 #include "catalog/pg_proc.h"
 #include "catalog/pg_type.h"
-#include "lib/stringinfo.h"
 #include "miscadmin.h"
 #include "parser/parse_type.h"
 #include "utils/builtins.h"
@@ -72,17 +70,17 @@ regprocin(PG_FUNCTION_ARGS)
 		strspn(pro_name_or_oid, "0123456789") == strlen(pro_name_or_oid))
 	{
 		result = DatumGetObjectId(DirectFunctionCall1(oidin,
-									  CStringGetDatum(pro_name_or_oid)));
+										  CStringGetDatum(pro_name_or_oid)));
 		PG_RETURN_OID(result);
 	}
 
 	/* Else it's a name, possibly schema-qualified */
 
 	/*
-	 * In bootstrap mode we assume the given name is not schema-qualified,
-	 * and just search pg_proc for a unique match.	This is needed for
-	 * initializing other system catalogs (pg_namespace may not exist yet,
-	 * and certainly there are no schemas other than pg_catalog).
+	 * In bootstrap mode we assume the given name is not schema-qualified, and
+	 * just search pg_proc for a unique match.	This is needed for
+	 * initializing other system catalogs (pg_namespace may not exist yet, and
+	 * certainly there are no schemas other than pg_catalog).
 	 */
 	if (IsBootstrapProcessingMode())
 	{
@@ -92,13 +90,13 @@ regprocin(PG_FUNCTION_ARGS)
 		SysScanDesc sysscan;
 		HeapTuple	tuple;
 
-		ScanKeyEntryInitialize(&skey[0], 0x0,
-							   (AttrNumber) Anum_pg_proc_proname,
-							   (RegProcedure) F_NAMEEQ,
-							   CStringGetDatum(pro_name_or_oid));
+		ScanKeyInit(&skey[0],
+					Anum_pg_proc_proname,
+					BTEqualStrategyNumber, F_NAMEEQ,
+					CStringGetDatum(pro_name_or_oid));
 
-		hdesc = heap_openr(ProcedureRelationName, AccessShareLock);
-		sysscan = systable_beginscan(hdesc, ProcedureNameNspIndex, true,
+		hdesc = heap_open(ProcedureRelationId, AccessShareLock);
+		sysscan = systable_beginscan(hdesc, ProcedureNameArgsNspIndexId, true,
 									 SnapshotNow, 1, skey);
 
 		while (HeapTupleIsValid(tuple = systable_getnext(sysscan)))
@@ -114,7 +112,7 @@ regprocin(PG_FUNCTION_ARGS)
 		if (matches == 0)
 			ereport(ERROR,
 					(errcode(ERRCODE_UNDEFINED_FUNCTION),
-				  errmsg("function \"%s\" does not exist", pro_name_or_oid)));
+				 errmsg("function \"%s\" does not exist", pro_name_or_oid)));
 
 		else if (matches > 1)
 			ereport(ERROR,
@@ -126,8 +124,8 @@ regprocin(PG_FUNCTION_ARGS)
 	}
 
 	/*
-	 * Normal case: parse the name into components and see if it matches
-	 * any pg_proc entries in the current search path.
+	 * Normal case: parse the name into components and see if it matches any
+	 * pg_proc entries in the current search path.
 	 */
 	names = stringToQualifiedNameList(pro_name_or_oid, "regprocin");
 	clist = FuncnameGetCandidates(names, -1);
@@ -173,9 +171,9 @@ regprocout(PG_FUNCTION_ARGS)
 		char	   *proname = NameStr(procform->proname);
 
 		/*
-		 * In bootstrap mode, skip the fancy namespace stuff and just
-		 * return the proc name.  (This path is only needed for debugging
-		 * output anyway.)
+		 * In bootstrap mode, skip the fancy namespace stuff and just return
+		 * the proc name.  (This path is only needed for debugging output
+		 * anyway.)
 		 */
 		if (IsBootstrapProcessingMode())
 			result = pstrdup(proname);
@@ -188,7 +186,7 @@ regprocout(PG_FUNCTION_ARGS)
 			 * Would this proc be found (uniquely!) by regprocin? If not,
 			 * qualify it.
 			 */
-			clist = FuncnameGetCandidates(makeList1(makeString(proname)), -1);
+			clist = FuncnameGetCandidates(list_make1(makeString(proname)), -1);
 			if (clist != NULL && clist->next == NULL &&
 				clist->oid == proid)
 				nspname = NULL;
@@ -259,15 +257,15 @@ regprocedurein(PG_FUNCTION_ARGS)
 		strspn(pro_name_or_oid, "0123456789") == strlen(pro_name_or_oid))
 	{
 		result = DatumGetObjectId(DirectFunctionCall1(oidin,
-									  CStringGetDatum(pro_name_or_oid)));
+										  CStringGetDatum(pro_name_or_oid)));
 		PG_RETURN_OID(result);
 	}
 
 	/*
-	 * Else it's a name and arguments.  Parse the name and arguments, look
-	 * up potential matches in the current namespace search list, and scan
-	 * to see which one exactly matches the given argument types.  (There
-	 * will not be more than one match.)
+	 * Else it's a name and arguments.  Parse the name and arguments, look up
+	 * potential matches in the current namespace search list, and scan to see
+	 * which one exactly matches the given argument types.	(There will not be
+	 * more than one match.)
 	 *
 	 * XXX at present, this code will not work in bootstrap mode, hence this
 	 * datatype cannot be used for any system column that needs to receive
@@ -324,8 +322,8 @@ format_procedure(Oid procedure_oid)
 		initStringInfo(&buf);
 
 		/*
-		 * Would this proc be found (given the right args) by
-		 * regprocedurein? If not, we need to qualify it.
+		 * Would this proc be found (given the right args) by regprocedurein?
+		 * If not, we need to qualify it.
 		 */
 		if (FunctionIsVisible(procedure_oid))
 			nspname = NULL;
@@ -336,11 +334,11 @@ format_procedure(Oid procedure_oid)
 						 quote_qualified_identifier(nspname, proname));
 		for (i = 0; i < nargs; i++)
 		{
-			Oid			thisargtype = procform->proargtypes[i];
+			Oid			thisargtype = procform->proargtypes.values[i];
 
 			if (i > 0)
 				appendStringInfoChar(&buf, ',');
-			appendStringInfo(&buf, "%s", format_type_be(thisargtype));
+			appendStringInfoString(&buf, format_type_be(thisargtype));
 		}
 		appendStringInfoChar(&buf, ')');
 
@@ -422,17 +420,17 @@ regoperin(PG_FUNCTION_ARGS)
 		strspn(opr_name_or_oid, "0123456789") == strlen(opr_name_or_oid))
 	{
 		result = DatumGetObjectId(DirectFunctionCall1(oidin,
-									  CStringGetDatum(opr_name_or_oid)));
+										  CStringGetDatum(opr_name_or_oid)));
 		PG_RETURN_OID(result);
 	}
 
 	/* Else it's a name, possibly schema-qualified */
 
 	/*
-	 * In bootstrap mode we assume the given name is not schema-qualified,
-	 * and just search pg_operator for a unique match.	This is needed for
-	 * initializing other system catalogs (pg_namespace may not exist yet,
-	 * and certainly there are no schemas other than pg_catalog).
+	 * In bootstrap mode we assume the given name is not schema-qualified, and
+	 * just search pg_operator for a unique match.	This is needed for
+	 * initializing other system catalogs (pg_namespace may not exist yet, and
+	 * certainly there are no schemas other than pg_catalog).
 	 */
 	if (IsBootstrapProcessingMode())
 	{
@@ -442,13 +440,13 @@ regoperin(PG_FUNCTION_ARGS)
 		SysScanDesc sysscan;
 		HeapTuple	tuple;
 
-		ScanKeyEntryInitialize(&skey[0], 0x0,
-							   (AttrNumber) Anum_pg_operator_oprname,
-							   (RegProcedure) F_NAMEEQ,
-							   CStringGetDatum(opr_name_or_oid));
+		ScanKeyInit(&skey[0],
+					Anum_pg_operator_oprname,
+					BTEqualStrategyNumber, F_NAMEEQ,
+					CStringGetDatum(opr_name_or_oid));
 
-		hdesc = heap_openr(OperatorRelationName, AccessShareLock);
-		sysscan = systable_beginscan(hdesc, OperatorNameNspIndex, true,
+		hdesc = heap_open(OperatorRelationId, AccessShareLock);
+		sysscan = systable_beginscan(hdesc, OperatorNameNspIndexId, true,
 									 SnapshotNow, 1, skey);
 
 		while (HeapTupleIsValid(tuple = systable_getnext(sysscan)))
@@ -464,7 +462,7 @@ regoperin(PG_FUNCTION_ARGS)
 		if (matches == 0)
 			ereport(ERROR,
 					(errcode(ERRCODE_UNDEFINED_FUNCTION),
-				   errmsg("operator does not exist: %s", opr_name_or_oid)));
+					 errmsg("operator does not exist: %s", opr_name_or_oid)));
 		else if (matches > 1)
 			ereport(ERROR,
 					(errcode(ERRCODE_AMBIGUOUS_FUNCTION),
@@ -475,8 +473,8 @@ regoperin(PG_FUNCTION_ARGS)
 	}
 
 	/*
-	 * Normal case: parse the name into components and see if it matches
-	 * any pg_operator entries in the current search path.
+	 * Normal case: parse the name into components and see if it matches any
+	 * pg_operator entries in the current search path.
 	 */
 	names = stringToQualifiedNameList(opr_name_or_oid, "regoperin");
 	clist = OpernameGetCandidates(names, '\0');
@@ -522,9 +520,9 @@ regoperout(PG_FUNCTION_ARGS)
 		char	   *oprname = NameStr(operform->oprname);
 
 		/*
-		 * In bootstrap mode, skip the fancy namespace stuff and just
-		 * return the oper name.  (This path is only needed for debugging
-		 * output anyway.)
+		 * In bootstrap mode, skip the fancy namespace stuff and just return
+		 * the oper name.  (This path is only needed for debugging output
+		 * anyway.)
 		 */
 		if (IsBootstrapProcessingMode())
 			result = pstrdup(oprname);
@@ -536,7 +534,7 @@ regoperout(PG_FUNCTION_ARGS)
 			 * Would this oper be found (uniquely!) by regoperin? If not,
 			 * qualify it.
 			 */
-			clist = OpernameGetCandidates(makeList1(makeString(oprname)),
+			clist = OpernameGetCandidates(list_make1(makeString(oprname)),
 										  '\0');
 			if (clist != NULL && clist->next == NULL &&
 				clist->oid == oprid)
@@ -557,8 +555,7 @@ regoperout(PG_FUNCTION_ARGS)
 	else
 	{
 		/*
-		 * If OID doesn't match any pg_operator entry, return it
-		 * numerically
+		 * If OID doesn't match any pg_operator entry, return it numerically
 		 */
 		result = (char *) palloc(NAMEDATALEN);
 		snprintf(result, NAMEDATALEN, "%u", oprid);
@@ -600,12 +597,10 @@ Datum
 regoperatorin(PG_FUNCTION_ARGS)
 {
 	char	   *opr_name_or_oid = PG_GETARG_CSTRING(0);
-	Oid			result = InvalidOid;
+	Oid			result;
 	List	   *names;
 	int			nargs;
 	Oid			argtypes[FUNC_MAX_ARGS];
-	char		oprkind;
-	FuncCandidateList clist;
 
 	/* '0' ? */
 	if (strcmp(opr_name_or_oid, "0") == 0)
@@ -617,15 +612,15 @@ regoperatorin(PG_FUNCTION_ARGS)
 		strspn(opr_name_or_oid, "0123456789") == strlen(opr_name_or_oid))
 	{
 		result = DatumGetObjectId(DirectFunctionCall1(oidin,
-									  CStringGetDatum(opr_name_or_oid)));
+										  CStringGetDatum(opr_name_or_oid)));
 		PG_RETURN_OID(result);
 	}
 
 	/*
-	 * Else it's a name and arguments.  Parse the name and arguments, look
-	 * up potential matches in the current namespace search list, and scan
-	 * to see which one exactly matches the given argument types.  (There
-	 * will not be more than one match.)
+	 * Else it's a name and arguments.  Parse the name and arguments, look up
+	 * potential matches in the current namespace search list, and scan to see
+	 * which one exactly matches the given argument types.	(There will not be
+	 * more than one match.)
 	 *
 	 * XXX at present, this code will not work in bootstrap mode, hence this
 	 * datatype cannot be used for any system column that needs to receive
@@ -644,28 +639,13 @@ regoperatorin(PG_FUNCTION_ARGS)
 				 errmsg("too many arguments"),
 				 errhint("Provide two argument types for operator.")));
 
-	if (argtypes[0] == InvalidOid)
-		oprkind = 'l';
-	else if (argtypes[1] == InvalidOid)
-		oprkind = 'r';
-	else
-		oprkind = 'b';
+	result = OpernameGetOprid(names, argtypes[0], argtypes[1]);
 
-	clist = OpernameGetCandidates(names, oprkind);
-
-	for (; clist; clist = clist->next)
-	{
-		if (memcmp(clist->args, argtypes, 2 * sizeof(Oid)) == 0)
-			break;
-	}
-
-	if (clist == NULL)
+	if (!OidIsValid(result))
 		ereport(ERROR,
 				(errcode(ERRCODE_UNDEFINED_FUNCTION),
 				 errmsg("operator does not exist: %s", opr_name_or_oid)));
 
-	result = clist->oid;
-
 	PG_RETURN_OID(result);
 }
 
@@ -697,8 +677,8 @@ format_operator(Oid operator_oid)
 		initStringInfo(&buf);
 
 		/*
-		 * Would this oper be found (given the right args) by
-		 * regoperatorin? If not, we need to qualify it.
+		 * Would this oper be found (given the right args) by regoperatorin?
+		 * If not, we need to qualify it.
 		 */
 		if (!OperatorIsVisible(operator_oid))
 		{
@@ -728,8 +708,7 @@ format_operator(Oid operator_oid)
 	else
 	{
 		/*
-		 * If OID doesn't match any pg_operator entry, return it
-		 * numerically
+		 * If OID doesn't match any pg_operator entry, return it numerically
 		 */
 		result = (char *) palloc(NAMEDATALEN);
 		snprintf(result, NAMEDATALEN, "%u", operator_oid);
@@ -798,20 +777,20 @@ regclassin(PG_FUNCTION_ARGS)
 	/* Numeric OID? */
 	if (class_name_or_oid[0] >= '0' &&
 		class_name_or_oid[0] <= '9' &&
-	strspn(class_name_or_oid, "0123456789") == strlen(class_name_or_oid))
+		strspn(class_name_or_oid, "0123456789") == strlen(class_name_or_oid))
 	{
 		result = DatumGetObjectId(DirectFunctionCall1(oidin,
-									CStringGetDatum(class_name_or_oid)));
+										CStringGetDatum(class_name_or_oid)));
 		PG_RETURN_OID(result);
 	}
 
 	/* Else it's a name, possibly schema-qualified */
 
 	/*
-	 * In bootstrap mode we assume the given name is not schema-qualified,
-	 * and just search pg_class for a match.  This is needed for
-	 * initializing other system catalogs (pg_namespace may not exist yet,
-	 * and certainly there are no schemas other than pg_catalog).
+	 * In bootstrap mode we assume the given name is not schema-qualified, and
+	 * just search pg_class for a match.  This is needed for initializing
+	 * other system catalogs (pg_namespace may not exist yet, and certainly
+	 * there are no schemas other than pg_catalog).
 	 */
 	if (IsBootstrapProcessingMode())
 	{
@@ -820,13 +799,13 @@ regclassin(PG_FUNCTION_ARGS)
 		SysScanDesc sysscan;
 		HeapTuple	tuple;
 
-		ScanKeyEntryInitialize(&skey[0], 0x0,
-							   (AttrNumber) Anum_pg_class_relname,
-							   (RegProcedure) F_NAMEEQ,
-							   CStringGetDatum(class_name_or_oid));
+		ScanKeyInit(&skey[0],
+					Anum_pg_class_relname,
+					BTEqualStrategyNumber, F_NAMEEQ,
+					CStringGetDatum(class_name_or_oid));
 
-		hdesc = heap_openr(RelationRelationName, AccessShareLock);
-		sysscan = systable_beginscan(hdesc, ClassNameNspIndex, true,
+		hdesc = heap_open(RelationRelationId, AccessShareLock);
+		sysscan = systable_beginscan(hdesc, ClassNameNspIndexId, true,
 									 SnapshotNow, 1, skey);
 
 		if (HeapTupleIsValid(tuple = systable_getnext(sysscan)))
@@ -834,7 +813,7 @@ regclassin(PG_FUNCTION_ARGS)
 		else
 			ereport(ERROR,
 					(errcode(ERRCODE_UNDEFINED_TABLE),
-					 errmsg("relation \"%s\" does not exist", class_name_or_oid)));
+			   errmsg("relation \"%s\" does not exist", class_name_or_oid)));
 
 		/* We assume there can be only one match */
 
@@ -845,8 +824,8 @@ regclassin(PG_FUNCTION_ARGS)
 	}
 
 	/*
-	 * Normal case: parse the name into components and see if it matches
-	 * any pg_class entries in the current search path.
+	 * Normal case: parse the name into components and see if it matches any
+	 * pg_class entries in the current search path.
 	 */
 	names = stringToQualifiedNameList(class_name_or_oid, "regclassin");
 
@@ -881,9 +860,9 @@ regclassout(PG_FUNCTION_ARGS)
 		char	   *classname = NameStr(classform->relname);
 
 		/*
-		 * In bootstrap mode, skip the fancy namespace stuff and just
-		 * return the class name.  (This path is only needed for debugging
-		 * output anyway.)
+		 * In bootstrap mode, skip the fancy namespace stuff and just return
+		 * the class name.	(This path is only needed for debugging output
+		 * anyway.)
 		 */
 		if (IsBootstrapProcessingMode())
 			result = pstrdup(classname);
@@ -892,8 +871,7 @@ regclassout(PG_FUNCTION_ARGS)
 			char	   *nspname;
 
 			/*
-			 * Would this class be found by regclassin? If not, qualify
-			 * it.
+			 * Would this class be found by regclassin? If not, qualify it.
 			 */
 			if (RelationIsVisible(classid))
 				nspname = NULL;
@@ -967,17 +945,17 @@ regtypein(PG_FUNCTION_ARGS)
 		strspn(typ_name_or_oid, "0123456789") == strlen(typ_name_or_oid))
 	{
 		result = DatumGetObjectId(DirectFunctionCall1(oidin,
-									  CStringGetDatum(typ_name_or_oid)));
+										  CStringGetDatum(typ_name_or_oid)));
 		PG_RETURN_OID(result);
 	}
 
 	/* Else it's a type name, possibly schema-qualified or decorated */
 
 	/*
-	 * In bootstrap mode we assume the given name is not schema-qualified,
-	 * and just search pg_type for a match.  This is needed for
-	 * initializing other system catalogs (pg_namespace may not exist yet,
-	 * and certainly there are no schemas other than pg_catalog).
+	 * In bootstrap mode we assume the given name is not schema-qualified, and
+	 * just search pg_type for a match.  This is needed for initializing other
+	 * system catalogs (pg_namespace may not exist yet, and certainly there
+	 * are no schemas other than pg_catalog).
 	 */
 	if (IsBootstrapProcessingMode())
 	{
@@ -986,13 +964,13 @@ regtypein(PG_FUNCTION_ARGS)
 		SysScanDesc sysscan;
 		HeapTuple	tuple;
 
-		ScanKeyEntryInitialize(&skey[0], 0x0,
-							   (AttrNumber) Anum_pg_type_typname,
-							   (RegProcedure) F_NAMEEQ,
-							   CStringGetDatum(typ_name_or_oid));
+		ScanKeyInit(&skey[0],
+					Anum_pg_type_typname,
+					BTEqualStrategyNumber, F_NAMEEQ,
+					CStringGetDatum(typ_name_or_oid));
 
-		hdesc = heap_openr(TypeRelationName, AccessShareLock);
-		sysscan = systable_beginscan(hdesc, TypeNameNspIndex, true,
+		hdesc = heap_open(TypeRelationId, AccessShareLock);
+		sysscan = systable_beginscan(hdesc, TypeNameNspIndexId, true,
 									 SnapshotNow, 1, skey);
 
 		if (HeapTupleIsValid(tuple = systable_getnext(sysscan)))
@@ -1011,8 +989,8 @@ regtypein(PG_FUNCTION_ARGS)
 	}
 
 	/*
-	 * Normal case: invoke the full parser to deal with special cases such
-	 * as array syntax.
+	 * Normal case: invoke the full parser to deal with special cases such as
+	 * array syntax.
 	 */
 	parseTypeString(typ_name_or_oid, &result, &typmod);
 
@@ -1044,9 +1022,9 @@ regtypeout(PG_FUNCTION_ARGS)
 		Form_pg_type typeform = (Form_pg_type) GETSTRUCT(typetup);
 
 		/*
-		 * In bootstrap mode, skip the fancy namespace stuff and just
-		 * return the type name.  (This path is only needed for debugging
-		 * output anyway.)
+		 * In bootstrap mode, skip the fancy namespace stuff and just return
+		 * the type name.  (This path is only needed for debugging output
+		 * anyway.)
 		 */
 		if (IsBootstrapProcessingMode())
 		{
@@ -1090,6 +1068,23 @@ regtypesend(PG_FUNCTION_ARGS)
 }
 
 
+/*
+ * text_regclass: convert text to regclass
+ */
+Datum
+text_regclass(PG_FUNCTION_ARGS)
+{
+	text	   *relname = PG_GETARG_TEXT_P(0);
+	Oid			result;
+	RangeVar   *rv;
+
+	rv = makeRangeVarFromNameList(textToQualifiedNameList(relname));
+	result = RangeVarGetRelid(rv, false);
+
+	PG_RETURN_OID(result);
+}
+
+
 /*
  * Given a C string, parse it into a qualified-name list.
  */
@@ -1099,7 +1094,7 @@ stringToQualifiedNameList(const char *string, const char *caller)
 	char	   *rawname;
 	List	   *result = NIL;
 	List	   *namelist;
-	List	   *l;
+	ListCell   *l;
 
 	/* We need a modifiable copy of the input string. */
 	rawname = pstrdup(string);
@@ -1122,7 +1117,7 @@ stringToQualifiedNameList(const char *string, const char *caller)
 	}
 
 	pfree(rawname);
-	freeList(namelist);
+	list_free(namelist);
 
 	return result;
 }
@@ -1259,7 +1254,7 @@ parseNameAndArgTypes(const char *string, const char *caller,
 			*ptr2 = '\0';
 		}
 
-		if (allowNone && strcasecmp(typename, "none") == 0)
+		if (allowNone && pg_strcasecmp(typename, "none") == 0)
 		{
 			/* Special case for NONE */
 			typeid = InvalidOid;
