@@ -1,4 +1,4 @@
-/* $PostgreSQL: pgsql/src/interfaces/ecpg/preproc/ecpg.c,v 1.91.4.1 2007/08/29 13:59:04 meskes Exp $ */
+/* $PostgreSQL: pgsql/src/interfaces/ecpg/preproc/ecpg.c,v 1.94.2.1 2007/08/29 13:58:34 meskes Exp $ */
 
 /* New main for ecpg, the PostgreSQL embedded SQL precompiler. */
 /* (C) Michael Meskes <meskes@postgresql.org> Feb 5th, 1998 */
@@ -43,7 +43,7 @@ help(const char *progname)
 	printf("  -c             automatically generate C code from embedded SQL code;\n"
 		   "                 currently this works for EXEC SQL TYPE\n");
 	printf("  -C MODE        set compatibility mode;\n"
-		   "                 MODE may be one of \"INFORMIX\", \"INFORMIX_SE\"\n");
+	  "                 MODE may be one of \"INFORMIX\", \"INFORMIX_SE\"\n");
 #ifdef YYDEBUG
 	printf("  -d             generate parser debug output\n");
 #endif
@@ -413,7 +413,7 @@ main(int argc, char *const argv[])
 					fprintf(yyout, "/* End of automatic include section */\n");
 				}
 
-				fprintf(yyout, "#line 1 \"%s\"\n", input_filename);
+				output_line_number();
 
 				/* and parse the source */
 				yyparse();
@@ -426,10 +426,10 @@ main(int argc, char *const argv[])
 					if (!(ptr->opened))
 					{
 						/*
-						 * Does not really make sense to declare a cursor
-						 * but not open it
+						 * Does not really make sense to declare a cursor but
+						 * not open it
 						 */
-						snprintf(errortext, sizeof(errortext), "cursor `%s´ has been declared but ot opened\n", ptr->name);
+						snprintf(errortext, sizeof(errortext), "cursor \"%s\" has been declared but not opened\n", ptr->name);
 						mmerror(PARSE_ERROR, ET_WARNING, errortext);
 					}
 					ptr = ptr->next;
