@@ -3,14 +3,14 @@
  * dirmod.c
  *	  rename/unlink()
  *
- * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2006, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *	These are replacement versions of unlink and rename that work on
  *	Win32 (NT, Win2k, XP).	replace() doesn't work on Win95/98/Me.
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/port/dirmod.c,v 1.34 2004/12/31 22:03:53 pgsql Exp $
+ *	  $PostgreSQL: pgsql/src/port/dirmod.c,v 1.44.2.4 2009/09/13 18:32:35 heikki Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -40,6 +40,7 @@
 #endif
 #endif
 
+
 #ifndef FRONTEND
 
 /*
@@ -52,8 +53,7 @@
 #define palloc(sz)		pgport_palloc(sz)
 #define pstrdup(str)	pgport_pstrdup(str)
 #endif
-
-#else /* FRONTEND */
+#else							/* FRONTEND */
 
 /*
  *	In frontend, fake palloc behavior with these
@@ -72,7 +72,7 @@ fe_palloc(Size size)
 
 	if ((res = malloc(size)) == NULL)
 	{
-		fprintf(stderr, gettext("out of memory\n"));
+		fprintf(stderr, _("out of memory\n"));
 		exit(1);
 	}
 	return res;
@@ -85,7 +85,7 @@ fe_pstrdup(const char *string)
 
 	if ((res = strdup(string)) == NULL)
 	{
-		fprintf(stderr, gettext("out of memory\n"));
+		fprintf(stderr, _("out of memory\n"));
 		exit(1);
 	}
 	return res;
@@ -98,13 +98,12 @@ fe_repalloc(void *pointer, Size size)
 
 	if ((res = realloc(pointer, size)) == NULL)
 	{
-		fprintf(stderr, gettext("out of memory\n"));
+		fprintf(stderr, _("out of memory\n"));
 		exit(1);
 	}
 	return res;
 }
-
-#endif /* FRONTEND */
+#endif   /* FRONTEND */
 
 
 #if defined(WIN32) || defined(__CYGWIN__)
@@ -117,40 +116,45 @@ pgrename(const char *from, const char *to)
 {
 	int			loops = 0;
 
-	/* Is this loop even necessary now that we have win32_open()?  */
+	/*
+	 * We need to loop because even though PostgreSQL uses flags that allow
+	 * rename while the file is open, other applications might have the file
+	 * open without those flags.  However, we won't wait indefinitely for
+	 * someone else to close the file, as the caller might be holding locks
+	 * and blocking other backends.
+	 */
 #if defined(WIN32) && !defined(__CYGWIN__)
 	while (!MoveFileEx(from, to, MOVEFILE_REPLACE_EXISTING))
+#else
+	while (rename(from, to) < 0)
 #endif
-#ifdef __CYGWIN__
-		while (rename(from, to) < 0)
-#endif
-		{
+	{
 #if defined(WIN32) && !defined(__CYGWIN__)
-			if (GetLastError() != ERROR_ACCESS_DENIED)
-#endif
-#ifdef __CYGWIN__
-				if (errno != EACCES)
-#endif
-					/* set errno? */
-					return -1;
-			pg_usleep(100000);	/* us */
-			if (loops == 30)
-#ifndef FRONTEND
-				elog(LOG, "could not rename \"%s\" to \"%s\", continuing to try",
-					 from, to);
-#else
-				fprintf(stderr, "could not rename \"%s\" to \"%s\", continuing to try\n",
-						from, to);
-#endif
-			loops++;
-		}
+		DWORD		err = GetLastError();
 
-	if (loops > 30)
-#ifndef FRONTEND
-		elog(LOG, "completed rename of \"%s\" to \"%s\"", from, to);
+		_dosmaperr(err);
+
+		/*
+		 * Modern NT-based Windows versions return ERROR_SHARING_VIOLATION
+		 * if another process has the file open without FILE_SHARE_DELETE.
+		 * ERROR_LOCK_VIOLATION has also been seen with some anti-virus
+		 * software. This used to check for just ERROR_ACCESS_DENIED, so
+		 * presumably you can get that too with some OS versions. We don't
+		 * expect real permission errors where we currently use rename().
+		 */
+		if (err != ERROR_ACCESS_DENIED &&
+			err != ERROR_SHARING_VIOLATION &&
+			err != ERROR_LOCK_VIOLATION)
+			return -1;
 #else
-		fprintf(stderr, "completed rename of \"%s\" to \"%s\"\n", from, to);
+		if (errno != EACCES)
+			return -1;
 #endif
+
+		if (++loops > 100)		/* time out after 10 sec */
+			return -1;
+		pg_usleep(100000);		/* us */
+	}
 	return 0;
 }
 
@@ -163,35 +167,26 @@ pgunlink(const char *path)
 {
 	int			loops = 0;
 
-	/* Is this loop even necessary now that we have win32_open()?  */
+	/*
+	 * We need to loop because even though PostgreSQL uses flags that allow
+	 * unlink while the file is open, other applications might have the file
+	 * open without those flags.  However, we won't wait indefinitely for
+	 * someone else to close the file, as the caller might be holding locks
+	 * and blocking other backends.
+	 */
 	while (unlink(path))
 	{
 		if (errno != EACCES)
-			/* set errno? */
+			return -1;
+		if (++loops > 100)		/* time out after 10 sec */
 			return -1;
 		pg_usleep(100000);		/* us */
-		if (loops == 30)
-#ifndef FRONTEND
-			elog(LOG, "could not unlink \"%s\", continuing to try",
-				 path);
-#else
-			fprintf(stderr, "could not unlink \"%s\", continuing to try\n",
-					path);
-#endif
-		loops++;
 	}
-
-	if (loops > 30)
-#ifndef FRONTEND
-		elog(LOG, "completed unlink of \"%s\"", path);
-#else
-		fprintf(stderr, "completed unlink of \"%s\"\n", path);
-#endif
 	return 0;
 }
 
 
-#ifdef WIN32	/* Cygwin has its own symlinks */
+#ifdef WIN32					/* Cygwin has its own symlinks */
 
 /*
  *	pgsymlink support:
@@ -235,7 +230,7 @@ pgsymlink(const char *oldpath, const char *newpath)
 	CreateDirectory(newpath, 0);
 	dirhandle = CreateFile(newpath, GENERIC_READ | GENERIC_WRITE,
 						   0, 0, OPEN_EXISTING,
-		   FILE_FLAG_OPEN_REPARSE_POINT | FILE_FLAG_BACKUP_SEMANTICS, 0);
+			   FILE_FLAG_OPEN_REPARSE_POINT | FILE_FLAG_BACKUP_SEMANTICS, 0);
 
 	if (dirhandle == INVALID_HANDLE_VALUE)
 		return -1;
@@ -261,13 +256,13 @@ pgsymlink(const char *oldpath, const char *newpath)
 						reparseBuf->PathBuffer, MAX_PATH);
 
 	/*
-	 * FSCTL_SET_REPARSE_POINT is coded differently depending on SDK
-	 * version; we use our own definition
+	 * FSCTL_SET_REPARSE_POINT is coded differently depending on SDK version;
+	 * we use our own definition
 	 */
 	if (!DeviceIoControl(dirhandle,
-						 CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 41, METHOD_BUFFERED, FILE_ANY_ACCESS),
+	 CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 41, METHOD_BUFFERED, FILE_ANY_ACCESS),
 						 reparseBuf,
-						 reparseBuf->ReparseDataLength + REPARSE_JUNCTION_DATA_BUFFER_HEADER_SIZE,
+	reparseBuf->ReparseDataLength + REPARSE_JUNCTION_DATA_BUFFER_HEADER_SIZE,
 						 0, 0, &len, 0))
 	{
 		LPSTR		msg;
@@ -275,15 +270,15 @@ pgsymlink(const char *oldpath, const char *newpath)
 		errno = 0;
 		FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
 					  NULL, GetLastError(),
-					  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
+					  MAKELANGID(LANG_ENGLISH, SUBLANG_DEFAULT),
 					  (LPSTR) & msg, 0, NULL);
 #ifndef FRONTEND
 		ereport(ERROR,
 				(errcode_for_file_access(),
-				 errmsg("Error setting junction for %s: %s",
+				 errmsg("could not set junction for \"%s\": %s",
 						nativeTarget, msg)));
 #else
-		fprintf(stderr, "Error setting junction for %s: %s\n",
+		fprintf(stderr, _("could not set junction for \"%s\": %s\n"),
 				nativeTarget, msg);
 #endif
 		LocalFree(msg);
@@ -297,10 +292,8 @@ pgsymlink(const char *oldpath, const char *newpath)
 
 	return 0;
 }
-
-#endif /* WIN32 */
-
-#endif /* defined(WIN32) || defined(__CYGWIN__) */
+#endif   /* WIN32 */
+#endif   /* defined(WIN32) || defined(__CYGWIN__) */
 
 
 /* We undefined this above, so we redefine it */
@@ -312,7 +305,7 @@ pgsymlink(const char *oldpath, const char *newpath)
 /*
  * fnames
  *
- * return a list of the names of objects in the argument directory 
+ * return a list of the names of objects in the argument directory
  */
 static char **
 fnames(char *path)
@@ -325,15 +318,24 @@ fnames(char *path)
 
 	dir = opendir(path);
 	if (dir == NULL)
+	{
+#ifndef FRONTEND
+		elog(WARNING, "could not open directory \"%s\": %m", path);
+#else
+		fprintf(stderr, _("could not open directory \"%s\": %s\n"),
+				path, strerror(errno));
+#endif
 		return NULL;
+	}
 
 	filenames = (char **) palloc(fnsize * sizeof(char *));
 
+	errno = 0;
 	while ((file = readdir(dir)) != NULL)
 	{
 		if (strcmp(file->d_name, ".") != 0 && strcmp(file->d_name, "..") != 0)
 		{
-			if (numnames+1 >= fnsize)
+			if (numnames + 1 >= fnsize)
 			{
 				fnsize *= 2;
 				filenames = (char **) repalloc(filenames,
@@ -341,6 +343,25 @@ fnames(char *path)
 			}
 			filenames[numnames++] = pstrdup(file->d_name);
 		}
+		errno = 0;
+	}
+#ifdef WIN32
+
+	/*
+	 * This fix is in mingw cvs (runtime/mingwex/dirent.c rev 1.4), but not in
+	 * released version
+	 */
+	if (GetLastError() == ERROR_NO_MORE_FILES)
+		errno = 0;
+#endif
+	if (errno)
+	{
+#ifndef FRONTEND
+		elog(WARNING, "could not read directory \"%s\": %m", path);
+#else
+		fprintf(stderr, _("could not read directory \"%s\": %s\n"),
+				path, strerror(errno));
+#endif
 	}
 
 	filenames[numnames] = NULL;
@@ -350,6 +371,7 @@ fnames(char *path)
 	return filenames;
 }
 
+
 /*
  *	fnames_cleanup
  *
@@ -366,6 +388,7 @@ fnames_cleanup(char **filenames)
 	pfree(filenames);
 }
 
+
 /*
  *	rmtree
  *
@@ -377,14 +400,15 @@ fnames_cleanup(char **filenames)
 bool
 rmtree(char *path, bool rmtopdir)
 {
-	char		filepath[MAXPGPATH];
+	char		pathbuf[MAXPGPATH];
+	char	   *filepath;
 	char	  **filenames;
 	char	  **filename;
 	struct stat statbuf;
 
 	/*
-	 * we copy all the names out of the directory before we start
-	 * modifying it.
+	 * we copy all the names out of the directory before we start modifying
+	 * it.
 	 */
 	filenames = fnames(path);
 
@@ -392,22 +416,21 @@ rmtree(char *path, bool rmtopdir)
 		return false;
 
 	/* now we have the names we can start removing things */
+	filepath = pathbuf;
 
 	for (filename = filenames; *filename; filename++)
 	{
 		snprintf(filepath, MAXPGPATH, "%s/%s", path, *filename);
 
-		if (stat(filepath, &statbuf) != 0)
-		{
-			fnames_cleanup(filenames);
-			return false;
-		}
+		if (lstat(filepath, &statbuf) != 0)
+			goto report_and_fail;
 
 		if (S_ISDIR(statbuf.st_mode))
 		{
 			/* call ourselves recursively for a directory */
 			if (!rmtree(filepath, true))
 			{
+				/* we already reported the error */
 				fnames_cleanup(filenames);
 				return false;
 			}
@@ -415,22 +438,65 @@ rmtree(char *path, bool rmtopdir)
 		else
 		{
 			if (unlink(filepath) != 0)
-			{
-				fnames_cleanup(filenames);
-				return false;
-			}
+				goto report_and_fail;
 		}
 	}
 
 	if (rmtopdir)
 	{
-		if (rmdir(path) != 0)
-		{
-			fnames_cleanup(filenames);
-			return false;
-		}
+		filepath = path;
+		if (rmdir(filepath) != 0)
+			goto report_and_fail;
 	}
 
 	fnames_cleanup(filenames);
 	return true;
+
+report_and_fail:
+
+#ifndef FRONTEND
+	elog(WARNING, "could not remove file or directory \"%s\": %m", filepath);
+#else
+	fprintf(stderr, _("could not remove file or directory \"%s\": %s\n"),
+			filepath, strerror(errno));
+#endif
+	fnames_cleanup(filenames);
+	return false;
 }
+
+
+#if defined(WIN32) && !defined(__CYGWIN__)
+
+#undef stat
+
+/*
+ * The stat() function in win32 is not guaranteed to update the st_size
+ * field when run. So we define our own version that uses the Win32 API
+ * to update this field.
+ */
+int 
+pgwin32_safestat(const char *path, struct stat *buf)
+{
+	int r;
+	WIN32_FILE_ATTRIBUTE_DATA attr;
+
+	r = stat(path, buf);
+	if (r < 0)
+		return r;
+
+	if (!GetFileAttributesEx(path, GetFileExInfoStandard, &attr))
+	{
+		_dosmaperr(GetLastError());
+		return -1;
+	}
+
+	/*
+	 * XXX no support for large files here, but we don't do that in
+	 * general on Win32 yet.
+	 */
+	buf->st_size = attr.nFileSizeLow;
+
+	return 0;
+}
+
+#endif
